"""Tests for ``otter.grade``"""

import logging
import os
import pandas as pd
import pytest
import re
import shutil
import subprocess
from unittest import mock
import zipfile

from glob import glob

from otter.generate import main as generate
from otter.generate.utils import zip_folder
from otter.grade import main as grade
from otter.utils import loggers

from .utils import TestFileManager


FILE_MANAGER = TestFileManager("test/test-grade")


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    yield
    if cleanup_enabled:
        if os.path.exists("test/final_grades.csv"):
            os.remove("test/final_grades.csv")
        if os.path.exists("test/submission_pdfs"):
            shutil.rmtree("test/submission_pdfs")


def generate_autograder_zip(pdfs=False):
    generate(
        tests_dir = FILE_MANAGER.get_path("tests"), 
        requirements = FILE_MANAGER.get_path("requirements.txt"), 
        output_path = FILE_MANAGER.get_path("autograder.zip"),
        config = FILE_MANAGER.get_path("otter_config.json") if pdfs else None,
        no_environment = True,
    )
    with zipfile.ZipFile(FILE_MANAGER.get_path("autograder.zip"), "a") as zip_ref:
        zip_folder(zip_ref, os.getcwd(), exclude=[".git", "logo", "test", "dist", "build", "otter_grader.egg-info"])


@pytest.fixture(autouse=True, scope="module")
def create_docker_image():    
    subprocess.run(["make", "docker-grade-test"], check=True)

    shutil.copy("otter/grade/Dockerfile", "otter/grade/old-Dockerfile")
    with open("otter/grade/Dockerfile", "r+") as f:
        lines = f.readlines()

        idx = max([i if "ARG" in lines[i] else -1 for i in range(len(lines))])
        lines.insert(idx + 1, "ADD otter-grader /home/otter-grader\n")

        f.seek(0)
        f.write("".join(lines))

    generate_autograder_zip(pdfs=True)

    yield

    subprocess.run(["make", "cleanup-docker-grade-test"], check=True)

    if os.path.exists("otter/grade/old-Dockerfile"):
        os.remove("otter/grade/Dockerfile")
        shutil.move("otter/grade/old-Dockerfile", "otter/grade/Dockerfile")

    if os.path.isfile(FILE_MANAGER.get_path("autograder.zip")):
        os.remove(FILE_MANAGER.get_path("autograder.zip"))

    # prune images
    grade(prune=True, force=True)


@pytest.fixture
def expected_points():
    """
    Load in point values
    """
    test_points = {}
    for test_file in glob(FILE_MANAGER.get_path("tests/*.py")):
        env = {}
        with open(test_file) as f:
            exec(f.read(), env)

        test_points[env['test']['name']] = env['test']['points']

    return test_points


@pytest.mark.slow
@pytest.mark.docker
def test_docker():
    """
    Check that we have the right container installed and that docker is running
    """
    # use docker image inspect to see that the image is installed and tagged as otter-grader
    inspect = subprocess.run(
        ["docker", "image", "inspect", "otter-test"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    # assert that it didn't fail, it will fail if it is not installed
    assert len(inspect.stderr) == 0, inspect.stderr.decode("utf-8")


@pytest.mark.slow
@pytest.mark.docker
def test_timeout():
    """
    Check that the notebook `20s.ipynb` is killed due to exceeding the defined timeout.
    """
    with pytest.raises(Exception, match=r"Executing '[\w./-]*test/test-grade/timeout/20s\.ipynb'" \
            " in docker container failed! Exit code: 137"), loggers.level_context(logging.DEBUG):
        grade(
            path=FILE_MANAGER.get_path("timeout/"),
            output_dir="test/",
            autograder=FILE_MANAGER.get_path("autograder.zip"),
            containers=5,
            image="otter-test",
            timeout=35,
        )


@pytest.mark.slow
@pytest.mark.docker
def test_network(expected_points):
    """
    Check that the notebook `network.ipynb` is unable to do some network requests with disabled networking
    """
    with loggers.level_context(logging.DEBUG):
        grade(
            path = FILE_MANAGER.get_path("network/"),
            output_dir = "test/",
            autograder = FILE_MANAGER.get_path("autograder.zip"),
            containers = 5,
            image = "otter-test",
            pdfs = True,
            no_network=True,
        )

    df_test = pd.read_csv("test/final_grades.csv")

    # sort by filename
    df_test = df_test.sort_values("file").reset_index(drop=True)

    for _, row in df_test.iterrows():
        for test in expected_points:
            if row['file'] == 'network.ipynb' and ('q2' in test or 'q3' in test):
                assert row[test] == 0, "{} supposed to fail {} but passed".format(row["file"], test)
            else:
                assert row[test] == expected_points[test], "{} supposed to pass {} but failed".format(row["file"], test)


@pytest.mark.slow
@pytest.mark.docker
def test_notebooks_with_pdfs(expected_points):
    """
    Check that the example of 100 notebooks runs correctely locally.
    """
    # grade the 100 notebooks
    with loggers.level_context(logging.DEBUG):
        grade(
            path = FILE_MANAGER.get_path("notebooks/"), 
            output_dir = "test/",
            autograder = FILE_MANAGER.get_path("autograder.zip"),
            containers = 5,
            image = "otter-test",
            pdfs = True,
        )

    # read the output and expected output
    df_test = pd.read_csv("test/final_grades.csv")

    # sort by filename
    df_test = df_test.sort_values("file").reset_index(drop=True)
    df_test["failures"] = df_test["file"].apply(lambda x: [int(n) for n in re.split(r"\D+", x) if len(n) > 0])

    # add score sum cols for tests
    for test in expected_points:
        test_cols = [l for l in df_test.columns if bool(re.search(fr"\b{test}\b", l))]
        df_test[test] = df_test[test_cols].sum(axis=1)

    # check point values
    for _, row in df_test.iterrows():
        for test in expected_points:
            if int(re.sub(r"\D", "", test)) in row["failures"]:
                # q6.py has all_or_nothing set to False, so if the hidden tests fail you should get 2.5 points
                if "6H" in row["file"] and "q6" == test:
                    assert row[test] == 2.5, "{} supposed to fail {} but passed".format(row["file"], test)
                else:
                    assert row[test] == 0, "{} supposed to fail {} but passed".format(row["file"], test)
            else:
                assert row[test] == expected_points[test], "{} supposed to pass {} but failed".format(row["file"], test)

    assert os.path.exists("test/submission_pdfs"), "PDF folder is missing"

    # check that an pdf exists for each submission
    dir1_contents, dir2_contents = (
        [os.path.splitext(f)[0] for f in os.listdir(FILE_MANAGER.get_path("notebooks/")) if not (os.path.isdir(os.path.join(FILE_MANAGER.get_path("notebooks/"), f)))],
        [os.path.splitext(f)[0] for f in os.listdir("test/submission_pdfs") if not (os.path.isdir(os.path.join("test/submission_pdfs", f)))],
    )
    assert sorted(dir1_contents) == sorted(dir2_contents), f"'{FILE_MANAGER.get_path('notebooks/')}' and 'test/submission_pdfs' have different contents"


def test_single_notebook_grade(expected_points):
    """
    Check that single notebook passed to grade returns percent.
    """
    data =  [{'q1': 2.0, 'q2':2.0, 'q3':2.0, 'q4':1.0, 'q6':5.0, \
                    'q2b':2.0, 'q7':1.0, 'percent_correct':1.0, 'file':'passesAll.ipynb'}]
    df = pd.DataFrame(data)
    notebook_path = FILE_MANAGER.get_path("notebooks/passesAll.ipynb")
    kw_expected = {
        "submissions_dir": mock.ANY,
        "num_containers": 1,
        "ext": 'ipynb',
        "no_kill": False,
        "output_path": 'test/',
        "zips": False,
        "image": 'otter-test',
        "pdfs": False,
        "timeout": None,
        "network": True
    }

    kws = {
        "path": notebook_path, 
        "output_dir": "test/",
        "autograder": notebook_path,
        "containers": 1,
        "image" : "otter-test",
        "pdfs" : False
    }

    with mock.patch("otter.grade.launch_grade") as mocked_launch_grade:
        mocked_launch_grade.return_value = [df]
        output = grade(**kws)
        mocked_launch_grade.assert_called_with(notebook_path, **kw_expected)
        assert output == 1.0



"""Tests for ``otter.run``"""

import copy
import json
import nbformat
import nbconvert
import os
import pytest
import re

from contextlib import nullcontext

from otter.run.run_autograder import main as run_autograder
from otter.run.run_autograder.utils import OtterRuntimeError
from otter.utils import NBFORMAT_VERSION

from .utils import delete_paths, TestFileManager


FILE_MANAGER = TestFileManager("test/test-run")


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    yield
    if cleanup_enabled:
        delete_paths([
            FILE_MANAGER.get_path("autograder/results/results.json"),
            FILE_MANAGER.get_path("autograder/results/results.pkl"),
            FILE_MANAGER.get_path("autograder/__init__.py"),
            FILE_MANAGER.get_path("autograder/submission/test"),
            FILE_MANAGER.get_path("autograder/submission/tests"),
            FILE_MANAGER.get_path("autograder/submission/__init__.py"),
            FILE_MANAGER.get_path("autograder/submission/.OTTER_LOG"),
            FILE_MANAGER.get_path("rmd-autograder/results/results.json"),
            FILE_MANAGER.get_path("rmd-autograder/results/results.pkl"),
            FILE_MANAGER.get_path("rmd-autograder/__init__.py"),
            FILE_MANAGER.get_path("rmd-autograder/submission/test"),
            FILE_MANAGER.get_path("rmd-autograder/submission/tests"),
            FILE_MANAGER.get_path("rmd-autograder/submission/__init__.py"),
            FILE_MANAGER.get_path("rmd-autograder/submission/.OTTER_LOG"),
        ])


@pytest.fixture(autouse=True)
def correct_cwd_on_exit():
    cwd = os.getcwd()
    yield
    os.chdir(cwd)


@pytest.fixture(scope="module")
def expected_results():
    return {
        "tests": [
            {
                "name": "Public Tests",
                "visibility": "visible",
                "output": "q1 results: All test cases passed!\n\nq2 results:\n    q2 - 1 result:\n        Trying:\n            negate(True)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 0\n        Failed example:\n            negate(True)\n        Expected:\n            False\n        Got:\n            True\n\n    q2 - 2 result:\n        Trying:\n            negate(False)\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 1\n        Failed example:\n            negate(False)\n        Expected:\n            True\n        Got:\n            False\n\nq3 results: All test cases passed!\n\nq4 results: All test cases passed!\n\nq6 results: All test cases passed!\n\nq7 results: All test cases passed!"
            },
            {
                "name": "q1",
                "score": 0.0,
                "max_score": 0.0,
                "visibility": "hidden",
                "output": "q1 results: All test cases passed!"
            },
            {
                "name": "q2",
                "score": 0,
                "max_score": 2.0,
                "visibility": "hidden",
                "output": "q2 results:\n    q2 - 1 result:\n        Trying:\n            negate(True)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 0\n        Failed example:\n            negate(True)\n        Expected:\n            False\n        Got:\n            True\n\n    q2 - 2 result:\n        Trying:\n            negate(False)\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 1\n        Failed example:\n            negate(False)\n        Expected:\n            True\n        Got:\n            False\n\n    q2 - 3 result:\n        Trying:\n            negate(\"\")\n        Expecting:\n            True\n        **********************************************************************\n        Line 2, in q2 2\n        Failed example:\n            negate(\"\")\n        Expected:\n            True\n        Got:\n            ''\n\n    q2 - 4 result:\n        Trying:\n            negate(1)\n        Expecting:\n            False\n        **********************************************************************\n        Line 2, in q2 3\n        Failed example:\n            negate(1)\n        Expected:\n            False\n        Got:\n            1"
            },
            {
                "name": "q3",
                "score": 2.0,
                "max_score": 2.0,
                "visibility": "hidden",
                "output": "q3 results: All test cases passed!"
            },
            {
                "name": "q4",
                "score": 1.0,
                "max_score": 1.0,
                "visibility": "hidden",
                "output": "q4 results: All test cases passed!"
            },
            {
                "name": "q6",
                "score": 2.5,
                "max_score": 5.0,
                "visibility": "hidden",
                "output": "q6 results:\n    q6 - 1 result:\n        Test case passed!\n\n    q6 - 2 result:\n        Trying:\n            fib = fiberator()\n        Expecting nothing\n        ok\n        Trying:\n            for _ in range(10):\n                print(next(fib))\n        Expecting:\n            0\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21\n            34\n        **********************************************************************\n        Line 3, in q6 1\n        Failed example:\n            for _ in range(10):\n                print(next(fib))\n        Expected:\n            0\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21\n            34\n        Got:\n            0\n            1\n            1\n            1\n            2\n            3\n            5\n            8\n            13\n            21"
            },
            {
                "name": "q7",
                "score": 1.0,
                "max_score": 1.0,
                "visibility": "hidden",
                "output": "q7 results: All test cases passed!"
            }
        ],
        "output": "Students are allowed 1 submissions every 1 days. You have 0 submissions in that period.",
    }


@pytest.fixture(scope="module")
def expected_rmd_results():
    return {
        "tests": [
            {
                "name": "Public Tests",
                "output": "q1 results: All test cases passed!",
                "visibility": "visible",
            },
            {
                "max_score": 5,
                "name": "q1",
                "output": "q1 results: All test cases passed!\nq1d message: congrats",
                "score": 5,
                "visibility": "hidden",
            },
        ],
    }


def get_expected_error_results(error):
    return {
        "score": 0,
        "stdout_visibility": "hidden",
        "tests": [
            {
                "name": "Autograder Error",
                "output": f"Otter encountered an error when grading this submission:\n\n{error}",
            },
        ],
    }


@pytest.fixture
def load_config():
    def load_config_file(rmd=False):
        dirname = "autograder" if not rmd else "rmd-autograder"
        with FILE_MANAGER.open(f"{dirname}/source/otter_config.json") as f:
            return json.load(f)
    return load_config_file


def test_notebook(load_config, expected_results):
    config = load_config()
    run_autograder(config['autograder_dir'])

    with FILE_MANAGER.open("autograder/results/results.json") as f:
        actual_results = json.load(f)

    assert actual_results == expected_results, \
        f"Actual results did not matched expected:\n{actual_results}"


def test_script(load_config, expected_results):
    config = load_config()
    nb_path = FILE_MANAGER.get_path("autograder/submission/fails2and6H.ipynb")
    nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)
    os.remove(nb_path)

    try:
        py, _ = nbconvert.export(nbconvert.PythonExporter, nb)

        # remove magic commands
        py = "\n".join(l for l in py.split("\n") if not l.startswith("get_ipython"))

        with FILE_MANAGER.open("autograder/submission/fails2and6H.py", "w+") as f:
            f.write(py)

        run_autograder(config['autograder_dir'])

        with FILE_MANAGER.open("autograder/results/results.json") as f:
            actual_results = json.load(f)

        assert actual_results == expected_results, \
            f"Actual results did not matched expected:\n{actual_results}"

    finally:
        delete_paths([FILE_MANAGER.get_path("autograder/submission/fails2and6H.py")])
        with open(nb_path, "w+") as f:
            nbformat.write(nb, f)


def test_assignment_name(load_config, expected_results):
    name = "hw01"
    config = load_config()
    nb_path = FILE_MANAGER.get_path("autograder/submission/fails2and6H.ipynb")
    orig_nb = nbformat.read(nb_path, as_version=NBFORMAT_VERSION)
    nb = copy.deepcopy(orig_nb)

    def perform_test(nb, expected_results, error=None, **kwargs):
        nbformat.write(nb, nb_path)

        cm = pytest.raises(OtterRuntimeError, match=re.escape(error)) if error is not None \
            else nullcontext()
        with cm:
            run_autograder(config["autograder_dir"], assignment_name = name, **kwargs)

        with FILE_MANAGER.open("autograder/results/results.json") as f:
            actual_results = json.load(f)

        assert actual_results == expected_results, \
            f"Actual results did not matched expected:\n{actual_results}"

    error_message_template = "Received submission for assignment '{got}' (this is assignment " \
            "'{want}')"

    try:
        # test with correct name
        nb["metadata"]["otter"] = {"assignment_name": name}
        perform_test(nb, expected_results)

        # test with wrong name
        bad_name = "lab01"
        error_message = error_message_template.format(got=bad_name, want=name)
        nb["metadata"]["otter"]["assignment_name"] = bad_name
        perform_test(nb, get_expected_error_results(error_message), error=error_message)

        # test with no name in nb
        error_message = error_message_template.format(got=None, want=name)
        nb["metadata"]["otter"].pop("assignment_name")
        perform_test(nb, get_expected_error_results(error_message), error=error_message)

    finally:
        delete_paths([nb_path])
        with open(nb_path, "w+") as f:
            nbformat.write(nb, f)


def test_rmd(load_config, expected_rmd_results):
    name = "hw01"
    config = load_config(True)
    rmd_path = FILE_MANAGER.get_path("rmd-autograder/submission/hw01.Rmd")
    with open(rmd_path) as f:
        orig_rmd = f.read()

    sub_name = lambda n: re.sub(r"assignment_name: \"\w+\"", f"assignment_name: \"{n}\"", orig_rmd)

    def perform_test(rmd, expected_results, error=None, **kwargs):
        with open(rmd_path, "w") as f:
            f.write(rmd)

        cm = pytest.raises(OtterRuntimeError, match=re.escape(error)) if error is not None \
            else nullcontext()
        with cm:
            run_autograder(config["autograder_dir"], assignment_name = name, **kwargs)

        with FILE_MANAGER.open("rmd-autograder/results/results.json") as f:
            actual_results = json.load(f)

        assert actual_results == expected_results, \
            f"Actual results did not matched expected:\n{actual_results}"

    error_message_template = "Received submission for assignment '{got}' (this is assignment " \
            "'{want}')"

    try:
        # test with correct name
        perform_test(orig_rmd, expected_rmd_results)

        # test with wrong name
        bad_name = "lab01"
        error_message = error_message_template.format(got=bad_name, want=name)
        perform_test(
            sub_name(bad_name), get_expected_error_results(error_message), error=error_message)

        # test with no name in Rmd
        error_message = error_message_template.format(got=None, want=name)
        perform_test(
            "\n".join([l for l in orig_rmd.split("\n") if not l.startswith("assignment_name: ")]),
            get_expected_error_results(error_message),
            error=error_message,
        )

    finally:
        delete_paths([rmd_path])
        with open(rmd_path, "w+") as f:
            f.write(orig_rmd)



"""Configurations for the autograder"""

import fica


class AutograderConfig(fica.Config):
    """
    Configurations for the autograder.
    """

    score_threshold = fica.Key(
        description="a score threshold for pass-fail assignments",
        default=None,
    )

    points_possible = fica.Key(
        description="a custom total score for the assignment; if unspecified the sum of question " \
            "point values is used.",
        default=None,
    )

    show_stdout = fica.Key(
        description="whether to display the autograding process stdout to students on Gradescope",
        default=False,
    )

    show_hidden = fica.Key(
        description="whether to display the results of hidden tests to students on Gradescope",
        default=False,
    )

    show_all_public = fica.Key(
        description="whether to display all test results if all tests are public tests",
        default=False,
    )

    seed = fica.Key(
        description="a random seed for intercell seeding",
        default=None,
    )

    seed_variable = fica.Key(
        description="a variable name to override with the seed",
        default=None,
    )

    grade_from_log = fica.Key(
        description="whether to re-assemble the student's environment from the log rather than " \
            "by re-executing their submission",
        default=False,
    )

    serialized_variables = fica.Key(
        description="a mapping of variable names to type strings for validating a deserialized " \
            "student environment",
        default=None,
    )

    pdf = fica.Key(
        description="whether to generate a PDF of the notebook when not using Gradescope " \
            "auto-upload",
        default=False,
    )

    token = fica.Key(
        description="a Gradescope token for uploading a PDF of the notebook",
        default=None,
    )

    course_id = fica.Key(
        description="a Gradescope course ID for uploading a PDF of the notebook",
        default="None",
    )

    assignment_id = fica.Key(
        description="a Gradescope assignment ID for uploading a PDF of the notebook",
        default="None",
    )

    filtering = fica.Key(
        description="whether the generated PDF should have cells filtered out",
        default=False,
    )

    pagebreaks = fica.Key(
        description="whether the generated PDF should have pagebreaks between filtered sections",
        default=False,
    )

    debug = fica.Key(
        description="whether to run the autograder in debug mode (without ignoring errors)",
        default=False,
    )

    autograder_dir = fica.Key(
        description="the directory in which autograding is taking place",
        default="/autograder",
    )

    lang = fica.Key(
        description="the language of the assignment; one of {'python', 'r'}",
        default="python",
        validator=fica.validators.choice(["python", "r"]),
    )

    miniconda_path = fica.Key(
        description="the path to the miniconda install directory",
        default="/root/miniconda3",
    )

    plugins = fica.Key(
        description="a list of plugin names and configuration details for grading",
        default=[],
    )

    logo = fica.Key(
        description="whether to print the Otter logo to stdout",
        default=True,
    )

    print_summary = fica.Key(
        description="whether to print the grading summary",
        default=True,
    )

    print_score = fica.Key(
        description="whether to print out the submission score in the grading summary",
        default=True,
    )

    zips = fica.Key(
        description="whether zip files are being graded",
        default=False,
    )

    log_level = fica.Key(
        description="a log level for logging messages; any value suitable for " \
            "``logging.Logger.setLevel``",
        default=None,
    )

    channel_priority_strict = fica.Key(
        description="whether to set conda's channel_priority config to strict in the setup.sh file",
        default=True,
    )

    assignment_name = fica.Key(
        description="a name for the assignment to ensure that students submit to the correct " \
            "autograder",
        default=None,
    )



"""Autograder configuration generator for Otter-Grader"""

import json
import os
import pathlib
import pkg_resources
import yaml
import zipfile

from glob import glob
from jinja2 import Template

from .token import APIClient
from .utils import zip_folder

from ..plugins import PluginCollection
from ..run.run_autograder.autograder_config import AutograderConfig
from ..utils import load_default_file


MINICONDA_INSTALL_URL = "https://repo.anaconda.com/miniconda/Miniconda3-py38_4.10.3-Linux-x86_64.sh"
OTTER_ENV_NAME = "otter-env"
OTTR_BRANCH = "v1.2.0"  # this should match a release tag on GitHub
TEMPLATE_DIR = pkg_resources.resource_filename(__name__, "templates")


LANGUAGE_BASED_CONFIGURATIONS = {
    "python": {
        "test_file_pattern": "*.py",
        "requirements_filename": "requirements.txt",
        "template_dir": os.path.join(TEMPLATE_DIR, "python"),
    },
    "r": {
        "test_file_pattern": "*.[Rr]",
        "requirements_filename": "requirements.R",
        "template_dir": os.path.join(TEMPLATE_DIR, "r"),
    },
}


def main(*, tests_dir="./tests", output_path="autograder.zip", config=None, no_config=False, 
         lang=None, requirements=None, no_requirements=False, overwrite_requirements=False, 
         environment=None, no_environment=False, username=None, password=None, token=None, files=[], 
         assignment=None, plugin_collection=None):
    """
    Runs Otter Generate

    Args:
        tests_dir (``str``): path to directory of test files for this assignment
        output_path (``str``): the path at which to write the output zip file
        config (``str``): path to an Otter configuration JSON file
        no_config (``bool``): disables auto-inclusion of Otter config file at ./otter_config.json
        lang (``str``): the language of the assignment; one of ``["python", "r"]``
        requirements (``str``): path to a Python or R requirements file for this assignment
        no_requirements (``bool``): disables auto-inclusion of requirements file at ./requirements.txt
        overwrite_requirements (``bool``): whether to overwrite the default requirements instead of
            adding to them
        environment (``str``): path to a conda environment file for this assignment
        no_environment (``bool``): whether ``./environment.yml`` should be automatically checked if 
            ``environment`` is unspecified
        username (``str``): a username for Gradescope for generating a token
        password (``str``): a password for Gradescope for generating a token
        token (``str``): a token for Gradescope
        files (``list[str]``): list of file paths to add to the zip file
        assignment (``otter.assign.assignment.Assignment``, optional): the assignment configurations
            if used with Otter Assign

    Raises:
        ``FileNotFoundError``: if the specified Otter configuration JSON file could not be found
        ``ValueError``: if the configurations specify a Gradescope course ID or assignment ID but not
            both
    """
    # read in otter_config.json
    if config is None and os.path.isfile("otter_config.json") and not no_config:
        config = "otter_config.json"

    if config is not None and not os.path.isfile(config):
        raise FileNotFoundError(f"Could not find otter configuration file {config}")

    if config:
        with open(config, encoding="utf-8") as f:
            otter_config = json.load(f)
    else:
        otter_config = {}

    # if an empty/null token is specified, delete it
    if "token" in otter_config and not otter_config["token"]:
        otter_config.pop("token")

    # ensure that a token is present if necessary
    if "token" not in otter_config and token is not None:
        otter_config["token"] = token

    elif "token" not in otter_config and "course_id" in otter_config and "assignment_id" in otter_config:
        client = APIClient()
        if username is not None and password is not None:
            client.log_in(username, password)
            token = client.token
        else:
            token = client.get_token()
        otter_config["token"] = token

    elif ("course_id" in otter_config) ^ ("assignment_id" in otter_config):
        raise ValueError(f"Otter config contains 'course_id' or 'assignment_id' but not both")

    ag_config = AutograderConfig(otter_config)

    # update language
    if lang is not None:
        ag_config.lang = lang
        otter_config["lang"] = lang

    if ag_config.lang not in LANGUAGE_BASED_CONFIGURATIONS.keys():
        raise ValueError(f"Invalid language specified: {ag_config.lang}")

    lang_config = LANGUAGE_BASED_CONFIGURATIONS[ag_config.lang]

    template_dir = lang_config["template_dir"]

    templates = {}
    for fn in os.listdir(template_dir):
        fp = os.path.join(template_dir, fn)
        if os.path.isfile(fp): # prevents issue w/ finding __pycache__ in template dirs
            with open(fp) as f:
                templates[fn] = Template(f.read())

    template_context = {
        "autograder_dir": ag_config.autograder_dir,
        "otter_env_name": OTTER_ENV_NAME,
        "miniconda_install_url": MINICONDA_INSTALL_URL,
        "ottr_branch": OTTR_BRANCH,
        "channel_priority_strict": ag_config.channel_priority_strict,
    }

    if plugin_collection is None:
        plugin_collection = PluginCollection(otter_config.get("plugins", []), None, {})

    else:
        plugin_collection.add_new_plugins(otter_config.get("plugins", []))

    plugin_collection.run("during_generate", otter_config, assignment)

    # open requirements if it exists
    with load_default_file(requirements, lang_config["requirements_filename"], 
                           default_disabled=no_requirements,) as reqs:
        template_context["other_requirements"] = reqs if reqs is not None else ""

    template_context["overwrite_requirements"] = overwrite_requirements

    # open environment if it exists
    # unlike requirements.txt, we will always overwrite, not append by default
    with load_default_file(environment, "environment.yml", default_disabled=no_environment) as env_contents:
        template_context["other_environment"] = env_contents
        if env_contents is not None:
            data = yaml.safe_load(env_contents)
            data['name'] = template_context["otter_env_name"]
            template_context["other_environment"] = yaml.safe_dump(data, default_flow_style=False)

    rendered = {}
    for fn, template in templates.items():
        rendered[fn] = template.render(**template_context)

    if os.path.exists(output_path):
        os.remove(output_path)

    with zipfile.ZipFile(output_path, mode="w") as zf:
        for fn, contents in rendered.items():
            zf.writestr(fn, contents)

        arc_test_dir = "tests"
        pattern = lang_config["test_file_pattern"]
        for file in glob(os.path.join(tests_dir, pattern)):
            zf.write(file, arcname=os.path.join(arc_test_dir, os.path.basename(file)))

        zf.writestr("otter_config.json", json.dumps(otter_config, indent=2))

        # copy files into tmp
        if len(files) > 0:
            for file in files:
                full_fp = os.path.abspath(file)
                if os.getcwd() not in full_fp:
                    raise ValueError(f"{file} is not in the working directory")

                relative_fp = pathlib.Path(full_fp).relative_to(pathlib.Path(os.getcwd()))
                if os.path.isfile(full_fp):
                    zf.write(file, arcname=os.path.join("files", relative_fp))
                elif os.path.isdir(full_fp):
                    zip_folder(zf, full_fp, prefix=os.path.join("files", os.path.split(relative_fp)[0]))
                else:
                    raise ValueError(f"Could not find file or directory '{full_fp}'")

    if assignment is not None:
        assignment._otter_config = otter_config



"""Utilities for Otter Generate"""

import os


def zip_folder(zf, path, prefix="", exclude=[]):
    """
    Recursively add the contents of a directory into a ``zipfile.ZipFile``.

    Args:
        zf (``zipfile.ZipFile``): the open zip file object to add to
        path (``str``): an absolute path to the directory to add to the zip file
        prefix (``str``, optional): a prefix to add to the basename of the path for the archive name
    """
    if not os.path.isabs(path):
        raise ValueError("'path' must be absolute path")

    parent_basename = os.path.basename(path)
    for file in os.listdir(path):
        if file in exclude:
            continue
        child_path = os.path.join(path, file)

        if os.path.isfile(child_path):
            arcname = os.path.join(prefix, parent_basename, file)
            zf.write(child_path, arcname=arcname)

        elif os.path.isdir(child_path):
            zip_folder(zf, child_path, prefix=os.path.join(prefix, parent_basename))



"""Execution of a Jupyter Notebook"""

import copy
import os
import tempfile

from contextlib import redirect_stdout, redirect_stderr
from IPython.display import display

try:
    from IPython.core.inputtransformer2 import TransformerManager
    _IPYTHON_7 = True
except ImportError:
    from IPython.core.inputsplitter import IPythonInputSplitter
    _IPYTHON_7 = False

from .transforms import create_collected_check_cell

from ..utils import id_generator


def execute_notebook(nb, check_results_list_name="check_results_secret", initial_env=None, 
                     ignore_errors=False, cwd=None, test_dir=None, seed=None, seed_variable=None):
    """
    Execute a notebook and return the global environment that results from execution.

    If ``ignore_errors`` is true, exceptions are swallowed.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to execute
        check_results_list_name (``str``, optional): the name of the list to collect check results in
        initial_env (``dict``, optional): an initial environment
        ignore_errors (``bool``, optional): whether exceptions should be ignored
        cwd (``str``, optional): working directory of execution to be appended to ``sys.path`` in 
            grading environment
        test_dir (``str``, optional): path to directory of tests in grading environment
        seed (``int``, optional): random seed for intercell seeding
        seed_variable (``str``, optional): a variable name to override with the seed

    Results:
        ``dict``: global environment resulting from executing all code of the input notebook
    """
    if initial_env:
        global_env = initial_env.copy()

    else:
        global_env = {}

    # add display from IPython
    global_env["display"] = display

    test_dir = test_dir if test_dir is not None else './tests'

    from ..check.notebook import Notebook
    with Notebook.grading_mode(tests_dir = test_dir if test_dir is not None else './tests') as check_results:
        # add dummy Notebook class so that we can collect results
        secret = id_generator()
        notebook_class_name = f"Notebook_{secret}"
        global_env[notebook_class_name] = Notebook

        source = ""

        if cwd:
            source = f"import sys\nsys.path.append(r\"{cwd}\")\n"
            exec(source, global_env)

        if seed is not None and seed_variable is None:
            import numpy as np
            import random
            global_env["np"] = np
            global_env["random"] = random

        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                if _IPYTHON_7:
                    isp = TransformerManager()

                else:
                    isp = IPythonInputSplitter(line_input_checker=False)

                try:
                    code_lines = []
                    cell_source_lines = cell['source']
                    # TODO: use otter.utils.get_source
                    source_is_str_bool = False
                    if isinstance(cell_source_lines, str):
                        source_is_str_bool = True
                        cell_source_lines = cell_source_lines.split('\n')

                    for line in cell_source_lines:
                        if not line.startswith('%') and  "interact(" not in line:
                            code_lines.append(line)
                            if source_is_str_bool:
                                code_lines.append('\n')

                    # TODO: move to helper function
                    if seed is not None:
                        if seed_variable is None:
                            cell_source = f"np.random.seed({seed})\nrandom.seed({seed})\n" + \
                                isp.transform_cell(''.join(code_lines))

                        else:
                            cell_source = f"{seed_variable} = {seed}\n" + \
                                isp.transform_cell(''.join(code_lines))

                    else:
                        cell_source = isp.transform_cell(''.join(code_lines))

                    with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
                        exec(cell_source, global_env)

                    source += cell_source

                except:
                    if not ignore_errors:
                        raise

            # add any required checks from this cell
            source += create_collected_check_cell(cell, notebook_class_name, test_dir)

        with tempfile.NamedTemporaryFile(mode="w", suffix=".py") as ntf:
            ntf.write(source)
            ntf.seek(0)

            try:
                cleaned_source = compile(source, filename=ntf.name, mode="exec")
                with open(os.devnull, 'w') as f, redirect_stdout(f), redirect_stderr(f):
                    exec(cleaned_source, global_env)

            except:
                if not ignore_errors:
                    raise

        # add the collected results to the global env
        global_env[check_results_list_name] = check_results

    return global_env



"""Assignment tests manager for Otter Assign"""

import os
import pprint
import re
import yaml

from dataclasses import dataclass
from jinja2 import Template
from typing import Any, Dict, List, Optional, Union

from .assignment import Assignment
from .question_config import QuestionConfig
from .solutions import remove_ignored_lines
from .utils import get_source, str_to_doctest

from ..test_files.abstract_test import OK_FORMAT_VARNAME, TestFile
from ..utils import NOTEBOOK_METADATA_KEY


BEGIN_TEST_CONFIG_REGEX = r'(?:.\s*=\s*)?""?"?\s*#\s*BEGIN\s*TEST\s*CONFIG'
END_TEST_CONFIG_REGEX = r'""?"?\s*#\s*END\s*TEST\s*CONFIG'

EXCEPTION_BASED_TEST_FILE_TEMPLATE = Template("""\
from otter.test_files import test_case

{{ OK_FORMAT_VARNAME }} = False

name = "{{ name }}"
points = {{ points }}

{% for tc in test_cases %}@test_case(points={{ tc.points }}, hidden={{ tc.hidden }}{% if tc.success_message is not none %}, 
    success_message="{{ tc.success_message }}"{% endif %}{% if tc.failure_message is not none %}, 
    failure_message="{{ tc.failure_message }}"{% endif %})
{{ tc.input }}
{% endfor %}
""")


@dataclass
class TestCase:
    """
    A dataclass representing a test case for a question.
    """

    input: str
    """the input of the test case"""

    output: str
    """the expected output of the test case"""

    hidden: bool
    """whether the test case is hidden"""

    points: Optional[Union[int, float]]
    """the point value of the test case"""

    success_message: Optional[str]
    """a message to show to the student if the test passes"""

    failure_message: Optional[str]
    """a message to show to the student if the test fails"""


class AssignmentTestsManager:
    """
    A class for creating and managing test cases for an assignment.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """

    assignment: Assignment
    """the assignment config"""

    _tests_by_question: Dict[str, List[TestCase]]
    """a dictionary mapping question names to lists of test cases"""

    _questions: Dict[str, QuestionConfig]
    """a dictionary mapping question names to ``QuestionConfig`` objects"""

    def __init__(self, assignment):
        self.assignment = assignment
        self._tests_by_question = {}
        self._questions = {}

    def any_public_tests(self, question):
        """
        Determine whether any of the test cases in the specified question are public.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``bool``: whether any of the test cases for the question are public
        """
        return any(not tc.hidden for tc in self._tests_by_question[question["name"]])

    def _add_test_case(self, question, test_case):
        """
        Track a test case for the specified question.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config
            test_case (``TestCase``): the test case to track
        """
        question_name = question["name"]
        self._questions[question_name] = question

        if question_name not in self._tests_by_question:
            self._tests_by_question[question_name] = []

        self._tests_by_question[question_name].append(test_case)

    def _parse_test_config(self, source: List[str]):
        """
        Parse test configurations from the test cell source.

        Args:
            source (``list[str]``): the source lines of the test cell

        Returns:
            ``tuple[dict, int | None]``: the parsed config and the index in ``source`` of the line
                at which the config ended, if any
        """
        config, i = {}, None
        if re.match(BEGIN_TEST_CONFIG_REGEX, source[0], flags=re.IGNORECASE):
            for i, line in enumerate(source):
                if re.match(END_TEST_CONFIG_REGEX, line, flags=re.IGNORECASE):
                    break
            config = yaml.full_load("\n".join(source[1:i]))
            assert isinstance(config, dict), f"Invalid test config in cell {source}"

        return config, i

    def read_test(self, cell, question):
        """
        Parse and track a test case from the provided cell for the specified question.

        Args:
            cell (``nbformat.NotebookNode``): a test cell
            question (``otter.assign.question_config.QuestionConfig``): the question config
        """
        source = get_source(cell)

        if source[0].lstrip().startswith("#"):
            hidden = bool(re.search(r"\bhidden\b", source[0], flags=re.IGNORECASE))
        else:
            hidden = False

        test_start_line = 0 if hidden else -1

        output = ''
        for o in cell['outputs']:
            output += ''.join(o.get('text', ''))
            results = o.get('data', {}).get('text/plain')
            if results and isinstance(results, list):
                output += results[0]
            elif results:
                output += results

        config, maybe_test_start_line = self._parse_test_config(source)
        test_start_line = maybe_test_start_line if maybe_test_start_line is not None else test_start_line

        hidden = config.get("hidden", hidden)
        points = config.get("points", None)
        success_message = config.get("success_message", None)
        failure_message = config.get("failure_message", None)

        test_source = "\n".join(remove_ignored_lines(get_source(cell)[test_start_line + 1:]))

        self._add_test_case(
            question,
            TestCase(test_source, output, hidden, points, success_message, failure_message),
        )

    def has_tests(self, question):
        """
        Determine whether the specified question has any test cases.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``bool``: whether the question has any test cases
        """
        return question.name in self._tests_by_question

    @staticmethod
    def _resolve_test_file_points(total_points, test_cases):
        """
        Validate and reformat the point values of the provided test cases taking into account the
        total points for the question.

        For Python, this is validation doesn't change the underlying test cases and only raises an
        error if the specified point values are invalid. This may not be the case for other
        languages, however.

        Args:
            total_points (``int | float | list[int | float | None] | None``): the value of the
                ``points`` configuration for the question
            test_cases (``list[TestCase]``): the test cases for the question

        Returns:
            ``list[TestCase]``: updated test cases
        """
        TestFile.resolve_test_file_points(total_points, test_cases)
        return test_cases

    def _create_test_file_info(self, question_name):
        """
        Create a ``dict`` containing the test file information for the question with the specified
        name.

        Args:
            question_name (``str``): the question name

        Returns:
            ``dict``: the test file info
        """
        question = self._questions[question_name]
        test_cases = self._tests_by_question[question_name]

        points = question.points
        if isinstance(points, dict):
            points = points.get('each', 1) * len(test_cases)
        elif isinstance(points, list):
            if len(points) != len(test_cases):
                raise ValueError(
                    f"Error in question {question.name}: length of 'points' is {len(points)} "
                    f"but there are {len(test_cases)} tests")

        # check for errors in resolving points
        test_cases = self._resolve_test_file_points(points, test_cases)

        return {
            "name": question.name,
            "points": points,
            "test_cases": test_cases,
        }

    @staticmethod
    def _create_ok_test_case(test_case: TestCase):
        """
        Create an OK-formatted test case for a test case object.

        Args:
            test (``TestCase``): the test case to convert to OK format

        Returns:
            ``dict``: the OK-formatted test case
        """
        code_lines = str_to_doctest(test_case.input.split('\n'), [])
        code_lines.append(test_case.output)

        ret = {
            'code': '\n'.join(code_lines),
            'hidden': test_case.hidden,
            'locked': False,
        }

        if test_case.points is not None:
            ret['points'] = test_case.points
        if test_case.success_message:
            ret['success_message'] = test_case.success_message
        if test_case.failure_message:
            ret['failure_message'] = test_case.failure_message

        return ret

    @classmethod
    def _create_ok_test_suite(cls, test_cases: List[TestCase]):
        """
        Create an OK-formatted test suite for a list of test cases.

        Args:
            test_cases (``list[TestCase]``): the test cases

        Returns:
            ``dict``: the OK-formatted test suite
        """
        return  {
            'cases': [cls._create_ok_test_case(tc) for tc in test_cases],
            'scored': True,
            'setup': '',
            'teardown': '',
            'type': 'doctest'
        }

    def _format_test(self, name, points, test_cases) -> Union[str, Dict[str, Any]]:
        """
        Format the test cases for a question based on the assignment config.

        Args:
            name (``str``): the name of the question
            points (``int | float | list[int | float | None] | None``): the ``points`` configuration
                from the question config
            test_cases (``list[TestCase]``): the test cases for the question

        Returns:
            ``dict | str``: the formatted test file
        """
        if self.assignment.tests.ok_format:
            test = {
                "name": name,
                "points": points,
                "suites": [self._create_ok_test_suite(test_cases)],
            }

        else:
            template_kwargs = {
                "name": name, 
                "points": points, 
                "test_cases": test_cases, 
                "OK_FORMAT_VARNAME": OK_FORMAT_VARNAME,
            }
            test = EXCEPTION_BASED_TEST_FILE_TEMPLATE.render(**template_kwargs)

        return test

    def write_tests(self, nb, test_dir, include_hidden=True, force_files=False):
        """
        Write all test files to a notebook's metadata or a tests directory.

        Args:
            nb (``nbformat.NotebookNode``): the notebook
            test_dir (``pathlib.Path | str``): the tests directory
            include_hidden (``bool``): whether to include hidden test cases
            force_files (``bool``): whether to force writing to test files (overriding the
                assignment config)
        """
        # TODO: move this notebook to the notebook metadata test classes
        if isinstance(nb, dict) and not self.assignment.tests.files and not force_files:
            if NOTEBOOK_METADATA_KEY not in nb["metadata"]:
                nb["metadata"][NOTEBOOK_METADATA_KEY] = {}
            nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"] = {}
            nb["metadata"][NOTEBOOK_METADATA_KEY][OK_FORMAT_VARNAME] = \
                self.assignment.tests.ok_format

        test_ext = ".py" if self.assignment.is_python else ".R"
        for test_name in self._tests_by_question.keys():
            test_info = self._create_test_file_info(test_name)
            test_path = os.path.join(test_dir, test_name + test_ext)

            if not include_hidden:
                test_info["test_cases"] = [tc for tc in test_info["test_cases"] if not tc.hidden]

            test = \
                self._format_test(test_info["name"], test_info["points"], test_info["test_cases"])

            if self.assignment.tests.files or force_files:
                with open(test_path, "w+") as f:
                    if isinstance(test, dict):
                        f.write(f"{OK_FORMAT_VARNAME} = True\n\ntest = ")
                        pprint.pprint(test, f, indent=4, width=200, depth=None)

                    else:
                        f.write(test)

            else:
                # TODO: move this notebook to the notebook metadata test classes
                nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"][test_info["name"]] = test

    def determine_question_point_value(self, question):
        """
        Determine the point value of a question using the question config and its test cases.

        Args:
            question (``otter.assign.question_config.QuestionConfig``): the question config

        Returns:
            ``int | float``: the point value of the question
        """
        test_cases = self._tests_by_question.get(question.name, [])
        if len(test_cases) == 0:
            if question.points is None and question.manual:
                raise ValueError(
                    f"Point value unspecified for question with no test cases: {question.name}")

            return question.points if question.points is not None else 1

        resolved_test_cases = TestFile.resolve_test_file_points(question.points, test_cases)
        points = round(sum(tc.points for tc in resolved_test_cases), 5)
        return int(points) if points % 1 == 0 else points



"""Utilities for Otter Assign"""

import copy
import datetime as dt
import json
import os
import pathlib
import re
import shutil

from glob import glob
from textwrap import indent

from ..execute import grade_notebook
from ..generate import main as generate_autograder
from ..utils import get_source, NOTEBOOK_METADATA_KEY


class EmptyCellException(Exception):
    """
    Exception for empty cells to indicate deletion
    """


class AssignNotebookFormatException(Exception):
    """
    """
    def __init__(self, message, question, cell_index, *args, **kwargs):
        message += " ("
        if question is not None:
            message += f"question { question.name }, "
        message += f"cell number { cell_index + 1 })"
        super().__init__(message, *args, **kwargs)


def get_notebook_language(nb):
    """
    Parse the notebook kernel language and return it as a string.

    Args:
        nb (``nbformat.NotebookNode``): the notebook

    Returns:
        ``str``: the name of the language as a lowercased string
    """
    return nb["metadata"]["kernelspec"]["language"].lower()


def is_ignore_cell(cell):
    """
    Returns whether the current cell should be ignored

    Args:
        cell (``nbformat.NotebookNode``): a notebook cell

    Returns:
        ``bool``: whether the cell is a ignored
    """
    source = get_source(cell)
    return bool(source and
        re.match(r"(##\s*ignore\s*##\s*|#\s*ignore\s*)", source[0], flags=re.IGNORECASE))


def is_cell_type(cell, cell_type):
    """
    Determine whether a cell is of a specific type.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        cell_type (``str``): the cell type to check for

    Returns:
        ``bool``: whether ``cell`` is of type ``cell_type``        
    """
    return cell["cell_type"] == cell_type


def remove_output(nb):
    """
    Remove all outputs from a notebook in-place.

    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb['cells']:
        if 'outputs' in cell:
            cell['outputs'] = []
        if 'execution_count' in cell:
            cell['execution_count'] = None


def remove_cell_ids(nb):
    """
    Remove all cell IDs from a notebook in-place.

    Args:
        nb (``nbformat.NotebookNode``): a notebook
    """
    for cell in nb['cells']:
        if 'id' in cell:
            cell.pop('id')


def lock(cell):
    """
    Make a cell non-editable and non-deletable in-place.

    Args:
        cell (``nbformat.NotebookNode``): cell to be locked
    """
    m = cell['metadata']
    m["editable"] = False
    m["deletable"] = False


def add_tag(cell, tag):
    """
    Add a tag to a cell, returning a copy.

    Args:
        cell (``nbformat.NotebookNode``): the cell to add a tag to
        tag (``str``): the tag to add

    Returns:
        ``nbformat.NotebookNode``: a copy of ``cell`` with the tag added
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"]:
        cell["metadata"]["tags"] = []
    cell["metadata"]["tags"].append(tag)
    return cell


def has_tag(cell, tag):
    """
    Determine whether a cell has a tag.

    Args:
        cell (``nbformat.NotebookNode``): the cell to check
        tag (``str``): the tag to check for

    Returns:
        ``nbformat.NotebookNode``: whether ``cell`` is tagged with ``tag``
    """
    return tag in cell["metadata"].get("tags", [])


def remove_tag(cell, tag):
    """
    Remove a tag from a cell, returning a copy.

    Args:
        cell (``nbformat.NotebookNode``): the cell to remove a tag from
        tag (``str``): the tag to remove

    Returns:
        ``nbformat.NotebookNode``: a copy of ``cell`` with the tag removed
    """
    cell = copy.deepcopy(cell)
    if "tags" not in cell["metadata"] or tag not in cell["metadata"]["tags"]:
        return cell
    cell["metadata"]["tags"].remove(tag)
    return cell


def str_to_doctest(code_lines, lines):
    """
    Convert a list of lines of Python code ``code_lines`` to the doctest format and appending the
    results to ``lines``.

    Args:
        code_lines (``list[str]``): the code to convert
        lines (``list[str]``): the list to append the converted lines to

    Returns:
        ``list[str]``: a pointer to ``lines``
    """
    if len(code_lines) == 0:
        return lines
    line = code_lines.pop(0)
    if line.startswith(" ") or line.startswith("\t"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif bool(re.match(r"^except[\s\w]*:", line)) or line.startswith("elif ") or \
            line.startswith("else:") or line.startswith("finally:"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    elif len(lines) > 0 and lines[-1].strip().endswith("\\"):
        return str_to_doctest(code_lines, lines + ["... " + line])
    else:
        return str_to_doctest(code_lines, lines + [">>> " + line])


def run_tests(nb_path, debug=False, seed=None, plugin_collection=None):
    """
    Grade a notebook and throw an error if it does not receive a perfect score.

    Args:
        nb_path (``pathlib.Path``): the path to the notebook to grade
        debug (``bool``, optional): whether to raise errors instead of ignoring them
        seed (``int``, optional): an RNG seed for notebook execution
        plugin_collection (``otter.plugins.PluginCollection``, optional): plugins to run while
            grading

    Raises:
        ``RuntimeError``: if the grade received by the notebook is not 100%
    """
    curr_dir = os.getcwd()
    os.chdir(nb_path.parent)

    results = grade_notebook(
        nb_path.name, tests_glob=glob(os.path.join("tests", "*.py")), cwd=os.getcwd(), 
    	test_dir=os.path.join(os.getcwd(), "tests"), ignore_errors = not debug, seed=seed,
        plugin_collection=plugin_collection
    )

    if results.total != results.possible:
        raise RuntimeError(f"Some autograder tests failed in the autograder notebook:\n" + \
            indent(results.summary(), '    '))

    os.chdir(curr_dir)


def write_otter_config_file(assignment):
    """
    Create an Otter configuration file (a ``.otter`` file) for students to use Otter tools,
    including saving environments and submitting to an Otter Service deployment, using assignment
    configurations. Writes the resulting file to the ``autograder`` and ``student`` subdirectories
    of ``assignment.result``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    config = {}

    config["notebook"] = assignment.master.name
    config["save_environment"] = assignment.save_environment
    config["ignore_modules"] = assignment.ignore_modules

    if assignment.variables:
        config["variables"] = assignment.variables

    config_name = assignment.master.stem + '.otter'
    with open(assignment.get_ag_path(config_name), "w+") as f:
        json.dump(config, f, indent=4)
    with open(assignment.get_stu_path(config_name), "w+") as f:
        json.dump(config, f, indent=4)


# TODO: update for new assign format
def run_generate_autograder(assignment, gs_username, gs_password, plugin_collection=None):
    """
    Run Otter Generate on the autograder directory to generate a Gradescope zip file. Relies on 
    configurations in ``assignment.generate``.

    Args:
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        gs_username (``str``): Gradescope username for token generation
        gs_password (``str``): Gradescope password for token generation
        plugin_collection (``otter.plugins.PluginCollection``, optional): a plugin collection to pass
            to Otter Generate
    """
    curr_dir = os.getcwd()
    os.chdir(str(assignment.get_ag_path()))

    # use temp tests dir
    test_dir = "tests"
    if assignment.is_python and not assignment.tests.files and assignment._temp_test_dir is None:
        raise RuntimeError("Failed to create temp tests directory for Otter Generate")

    elif assignment.is_python and not assignment.tests.files:
        test_dir = str(assignment._temp_test_dir)

    requirements = None
    if assignment.requirements:
        if assignment.is_r:
            requirements = 'requirements.R'
        else:
            requirements = 'requirements.txt'

    files = []
    if assignment.files:
        files += assignment.files

    if assignment.autograder_files:
        ag_dir = os.getcwd()
        os.chdir(curr_dir)
        output_dir = assignment.get_ag_path()

        # copy files
        for file in assignment.autograder_files:

            # if a directory, copy the entire dir
            if os.path.isdir(file):
                shutil.copytree(file, str(output_dir / os.path.basename(file)))

            else:
                # check that file is in subdir
                assert os.getcwd() in os.path.abspath(file), \
                    f"{file} is not in a subdirectory of the master notebook directory"
                file_path = pathlib.Path(file).resolve()
                rel_path = file_path.parent.relative_to(pathlib.Path(os.getcwd()))
                os.makedirs(output_dir / rel_path, exist_ok=True)
                shutil.copy(file, str(output_dir / rel_path))

        os.chdir(ag_dir)

        files += assignment.autograder_files

    otter_config = assignment.get_otter_config()
    if otter_config:
        # TODO: move this filename into a global variable somewhere and remove all of the places
        # it's hardcoded
        with open("otter_config.json", "w+") as f:
            json.dump(otter_config, f, indent=2)

    # TODO: change generate_autograder so that only necessary kwargs are needed
    timestamp = dt.datetime.now().strftime("%Y_%m_%dT%H_%M_%S_%f")
    notebook_name = assignment.master.stem
    output_path = f"{notebook_name}-autograder_{timestamp}.zip"
    generate_autograder(
        tests_dir=test_dir,
        output_path=output_path,
        config="otter_config.json" if otter_config else None,
        lang="python" if assignment.is_python else "r",
        requirements=requirements,
        overwrite_requirements=assignment.overwrite_requirements,
        environment="environment.yml" if assignment.environment else None,
        no_environment=False,
        username=gs_username,
        password=gs_password,
        files=files,
        plugin_collection=plugin_collection,
        assignment=assignment,
    )

    # clean up temp tests dir
    if assignment._temp_test_dir is not None:
        shutil.rmtree(str(assignment._temp_test_dir))

    os.chdir(curr_dir)


def add_assignment_name_to_notebook(nb, assignment):
    """
    Add the assignment name from the assignment config to the provided notebook's metadata in-place.

    If ``assignment`` has a name, the name is added to the notebook metadata, as
    ``nb["metadata"]["otter"]["assignment_name"]``.

    Args:
        nb (``nbformat.NotebookNode``): the notebook to add the name to
        assignment (``otter.assign.assignment.Assignment``): the assignment config
    """
    if assignment.name is not None:
        if NOTEBOOK_METADATA_KEY not in nb["metadata"]:
            nb["metadata"][NOTEBOOK_METADATA_KEY] = {}
        nb["metadata"][NOTEBOOK_METADATA_KEY]["assignment_name"] = assignment.name



"""Class for running tests from test files"""

import inspect

from ..test_files import create_test_file


class Checker:
    """
    A class for running and optionally tracking checks against test files.

    This class is not meant to be instantiated and is composed solely of class methods.
    """

    _track_results = False
    _test_files = []

    def __new__(cls, *args, **kwargs):
        raise NotImplementedError("The Checker class cannot be instantiated")

    @classmethod
    def enable_tracking(cls):
        """
        Enable the tracking of test results from calls to ``check``.
        """
        cls._track_results = True

    @classmethod
    def disable_tracking(cls):
        """
        Disable the tracking of test results from calls to ``check``.
        """
        cls._track_results = False

    @classmethod
    def get_results(cls):
        """
        Get a pointer to the list into which check results are being collected.
        """
        return cls._test_files

    @classmethod
    def clear_results(cls):
        """
        Overwrite the pointer to the check result collection list.

        Does not affect the original list, only overwrites the field in the ``Checker`` class that
        points to it.
        """
        cls._test_files = []

    @classmethod
    def check(cls, nb_or_test_path, test_name=None, global_env=None):
        """
        Checks a global environment against given test file. If ``global_env`` is ``None``, the global 
        environment of the calling frame is used; i.e., the following two calls are equivalent:

        .. code-block:: python

            check('tests/q1.py')
            check('tests/q1.py', global_env=globals())

        Args:
            nb_or_test_path (``str``): path to test file or notebook
            test_name (``str``, optional): the name of the test if a notebook metadata test
            global_env (``dict``, optional): a global environment resulting from the execution 
                of a python script or notebook

        Returns:
            ``otter.test_files.abstract_test.TestFile``: result of running the tests in the 
            given global environment
        """
        test = create_test_file(nb_or_test_path, test_name=test_name)

        if global_env is None:
            global_env = inspect.currentframe().f_back.f_globals

        test.run(global_env)

        if cls._track_results:
            cls._test_files.append(test)

        return test



"""
ottr test adapters for Otter Assign
"""

import re
import pprint
import yaml
import nbformat

from collections import namedtuple
from dataclasses import dataclass, replace
from typing import Optional, Union

from ..constants import TEST_REGEX, OTTR_TEST_NAME_REGEX, OTTR_TEST_FILE_TEMPLATE
from ..tests import write_test
from ..utils import get_source, lock

Test = namedtuple('Test', ['name', 'hidden', 'points', 'body', 'success_message', 'failure_message'])

@dataclass
class Test:

    name: str

    hidden: bool

    points: Union[int, float]

    body: str

    success_message: Optional[str]

    failure_message: Optional[str]


def read_test(cell, question, assignment, rmd=False):
    """
    Returns the contents of a test as a ``(name, hidden, body)`` named tuple

    Args:
        cell (``nbformat.NotebookNode``): a test cell
        question (``dict``): question metadata
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        rmd (``bool``, optional): whether the cell is from an Rmd file; if true, the first and last
            lines of ``cell``'s source are trimmed, since they should be backtick delimeters

    Returns:
        ``Test``: test named tuple
    """
    if rmd:
        source = get_source(cell)[1:-1]
    else:
        source = get_source(cell)
    hidden = bool(re.search("hidden", source[0], flags=re.IGNORECASE))
    lines = source[1:]
    test_name = None
    # for line in lines:
    #     match = re.match(OTTR_TEST_NAME_REGEX, line)
    #     if match:
    #         test_name = match.group(1)
    #         break
    # assert test_name is not None, f"Could not parse test name:\n{cell}"
    # TODO: hook up success_message and failure_message
    # TODO: add parsing for TEST CONFIG blocks
    return Test(test_name, hidden, None, '\n'.join(lines), "", "")

def gen_test_cell(question, tests, tests_dict, assignment):
    """
    Parses a list of test named tuples and creates a single test file. Adds this test file as a value
    to ``tests_dict`` with a key corresponding to the test's name, taken from ``question``. Returns
    a code cell that runs the check on this test.

    Args:
        question (``dict``): question metadata
        tests (``list`` of ``Test``): tests to be written
        tests_dict (``dict``): the tests for this assignment
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations

    Returns:
        ``nbformat.NotebookNode``: code cell calling ``ottr::check`` on this test
    """
    cell = nbformat.v4.new_code_cell()
    cell.source = ['. = ottr::check("tests/{}.R")'.format(question['name'])]

    points = question.get('points', len(tests))
    if isinstance(points, (int, float)):
        if points % len(tests) == 0:
            points = [points // len(tests) for _ in range(len(tests))]
        else:
            points = [points / len(tests) for _ in range(len(tests))]
    assert isinstance(points, list) and len(points) == len(tests), \
        f"Points for question {question['name']} could not be parsed:\n{points}"

    # update point values
    tests = [replace(tc, points=p) for tc, p in zip(tests, points)]
    test = gen_suite(question['name'], tests, points)

    tests_dict[question['name']] = test
    lock(cell)
    return cell

def gen_suite(name, tests, points):
    """
    Generates an R-formatted test file for ottr

    Args:
        name (``str``): the test name
        tests (``list`` of ``Test``): the test case named tuples that define this test file
        points (``float`` or ``int`` or ``list`` of ``float`` or ``int``): th points per question

    Returns:
        ``str``: the rendered R test file
    """
    template_data = {'name': name, 'test_cases': tests}
    return OTTR_TEST_FILE_TEMPLATE.render(**template_data)

def remove_hidden_tests_from_dir(nb, test_dir, assignment, use_files=True):
    """
    Rewrites test files in a directory to remove hidden tests

    Args:
        nb (``nbformat.NotebookNode``): the student notebook
        test_dir (``pathlib.Path``): path to test files directory
        assignment (``otter.assign.assignment.Assignment``): the assignment configurations
        use_files (``bool``, optional): ignored for R assignments
    """
    for f in test_dir.iterdir():
        if f.suffix != '.R':
            continue

        with open(f) as f2:
            test = f2.read()

        test = re.sub(r"    ottr::TestCase\$new\(\s*hidden = TRUE[\w\W]+?^    \),?", "", test, flags=re.MULTILINE)
        test = re.sub(r",(\s*  \))", r"\1", test, flags=re.MULTILINE)  # removes a trailing comma if present

        write_test({}, f, test, use_file=True)



"""Assignment configurations for Otter Assign"""

import fica
import os
import pathlib

from typing import Any, Dict, List, Optional, Union

from ..run.run_autograder.autograder_config import AutograderConfig
from ..utils import Loggable


# TODO: add detection/warnings/errors for when a user provides an invalid key? (to be added to fica)
class Assignment(fica.Config, Loggable):
    """
    Configurations for the assignment.
    """

    name: Optional[str] = fica.Key(
        description="a name for the assignment (to validate that students submit to the correct " \
            "autograder)",
        default=None,
    )

    requirements: Optional[str] = fica.Key(
        description="the path to a requirements.txt file or a list of packages",
        default=None,
    )

    overwrite_requirements: bool = fica.Key(
        description="whether to overwrite Otter's default requirement.txt in Otter Generate",
        default=False,
    )

    environment: Optional[str] = fica.Key(
        description="the path to a conda environment.yml file",
        default=None,
    )

    run_tests: bool = fica.Key(
        description="whether to run the assignment tests against the autograder notebook",
        default=True,
    )

    solutions_pdf: bool = fica.Key(
        description="whether to generate a PDF of the solutions notebook",
        default=False,
    )

    template_pdf: bool = fica.Key(
        description="whether to generate a filtered Gradescope assignment template PDF",
        default=False,
    )

    init_cell: bool = fica.Key(
        description="whether to include an Otter initialization cell in the output notebooks",
        default=True,
    )

    check_all_cell: bool = fica.Key(
        description="whether to include an Otter check-all cell in the output notebooks",
        default=False,
    )

    class ExportCellValue(fica.Config):

        instructions: str = fica.Key(
            description="additional submission instructions to include in the export cell",
            default="",
        )

        pdf: bool = fica.Key(
            description="whether to include a PDF of the notebook in the generated zip file",
            default=True,
        )

        filtering: bool = fica.Key(
            description="whether the generated PDF should be filtered",
            default=True,
        )

        force_save: bool = fica.Key(
            description="whether to force-save the notebook with JavaScript (only works in " \
                "classic notebook)",
            default=False,
        )

        run_tests: bool = fica.Key(
            description="whether to run student submissions against local tests during export",
            default=True,
        )

    export_cell: ExportCellValue = fica.Key(
        description="whether to include an Otter export cell in the output notebooks",
        subkey_container=ExportCellValue,
    )

    class SeedValue(fica.Config):

        variable: Optional[str] = fica.Key(
            description="a variable name to override with the autograder seed during grading",
            default=None,
        )

        autograder_value: Optional[int] = fica.Key(
            description="the value of the autograder seed",
            default=None,
        )

        student_value: Optional[int] = fica.Key(
            description="the value of the student seed",
            default=None,
        )

    seed: SeedValue = fica.Key(
        description="intercell seeding configurations",
        default=None,
        subkey_container=SeedValue,
    )

    generate: Union[bool, AutograderConfig] = fica.Key(
        description="grading configurations to be passed to Otter Generate as an " \
            "otter_config.json; if false, Otter Generate is disabled",
        default=False,
        subkey_container=AutograderConfig,
    )

    save_environment: bool = fica.Key(
        description="whether to save the student's environment in the log",
        default=False,
    )

    variables: Optional[Dict[str, str]] = fica.Key(
        description="a mapping of variable names to type strings for serializing environments",
        default=None,
    )

    ignore_modules: List[str] = fica.Key(
        description="a list of modules to ignore variables from during environment serialization",
        default=[],
    )

    files: List[str] = fica.Key(
        description="a list of other files to include in the output directories and autograder",
        default=[],
    )

    autograder_files: List[str] = fica.Key(
        description="a list of other files only to include in the autograder",
        default=[],
    )

    plugins: List[str] = fica.Key(
        description="a list of plugin names and configurations",
        default=[],
    )

    class TestsValue(fica.Config):

        files: bool = fica.Key(
            description="whether to store tests in separate files, instead of the notebook " \
                "metadata",
            default=False,
        )

        ok_format: bool = fica.Key(
            description="whether the test cases are in OK-format (instead of the exception-based " \
                "format)",
            default=True,
        )

        url_prefix: Optional[str] = fica.Key(
            description="a URL prefix for where test files can be found for student use",
            default=None,
        )

    tests: TestsValue = fica.Key(
        description="information about the structure and storage of tests",
        subkey_container=TestsValue,
        enforce_subkeys=True,
    )

    show_question_points: bool = fica.Key(
        description="whether to add the question point values to the last cell of each question",
        default=False,
    )

    runs_on: str = fica.Key(
        description= "the interpreter this notebook will be run on if different from the " \
            "default interpreter (one of {'default', 'colab', 'jupyterlite'})",
        default="default",
        validator=fica.validators.choice(["default", "colab", "jupyterlite"])
    )

    lang: Optional[str] = None
    """the language of the assignment"""

    master: pathlib.Path = None
    """the path to the master notebook"""

    result: pathlib.Path = None
    """the path to the output directory"""

    seed_required: bool = False
    """whether a seeding configuration is required for Otter Generate"""

    # TODO: rename this
    _temp_test_dir: Optional[str] = None
    """the path to a directory of test files for Otter Generate"""

    notebook_basename: Optional[str] = None
    """the basename of the master notebook file"""

    def __init__(self, user_config: Dict[str, Any] = {}, **kwargs) -> None:
        self._logger.debug(f"Initializing with config: {user_config}")
        super().__init__(user_config, **kwargs)

        # convert a boolean to a config object for self.generate if indicated
        if self.generate is True:
            self.generate = AutograderConfig()

    def update(self, user_config: Dict[str, Any]):
        self._logger.debug(f"Updating config: {user_config}")
        ret = super().update(user_config)
        if self.generate is True:
            self.generate = AutograderConfig()
        return ret

    @property
    def is_r(self):
        """
        Whether the language of the assignment is R
        """
        return self.lang == "r"

    @property
    def is_python(self):
        """
        Whether the language of the assignment is Python
        """
        return self.lang == "python"

    @property
    def is_rmd(self):
        """
        Whether the input file is an RMarkdown document
        """
        return self.master.suffix.lower() == ".rmd"

    @property
    def generate_enabled(self):
        """
        Whether Otter Generate is enabled for this assignment
        """
        return self.generate is not False

    def get_otter_config(self):
        """
        Get the contents of ``otter_config.json`` for this assignment.

        Returns:
            ``dict[str, object]``: the ``otter_config.json`` file as a ``dict``
        """
        if not self.generate_enabled:
            raise ValueError("Otter Generate is not configured for this assignment")

        otter_config = self.generate

        if self.is_r:
            otter_config.lang = "r"

        # TODO: move this config out of the assignment metadata and into the generate key
        if self.variables:
            otter_config.serialized_variables = str(self.variables)

        if self.name:
            otter_config.assignment_name = self.name

        return otter_config.get_user_config()

    @property
    def notebook_basename(self):
        """the basename of the notebook"""
        return os.path.basename(str(self.master))

    @property
    def ag_notebook_path(self):
        """the path to the autograder notebook"""
        return self.get_ag_path(self.notebook_basename)

    def get_ag_path(self, path=""):
        """
        Get the path to the autograder output directory or a file in that directory.

        Args:
            path (``str | pathlib.Path``): a path to append to the autograder output directory path

        Returns:
            ``pathlib.Path``: the path to the autograder directory or the specified file within it
        """
        return self.result / "autograder" / path

    def get_stu_path(self, path=""):
        """
        Get the path to the student output directory or a file in that directory.

        Args:
            path (``str | pathlib.Path``): a path to append to the student output directory path

        Returns:
            ``pathlib.Path``: the path to the student directory or the specified file within it
        """
        return self.result / "student" / path



"""Tests for ``otter.generate.token``"""

from unittest import mock

from otter.generate.token import APIClient


client = APIClient("token1")


# This method will be used by the mock to replace requests.post
def mocked_requests_post(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data, status_code):
            self.text = json_data
            self.status_code = status_code

        def json(self):
            return self.text

    return MockResponse({"token": "value1"}, 200)


@mock.patch("otter.generate.token.getpass.getpass")
@mock.patch.object(APIClient, "post")
@mock.patch('builtins.input', return_value='user')
def test_get_token(mocked_getpass, mocked_post, _):
    """
    Tests happy path for ``otter.generate.token.APIClient.get_token``, which asks the user for their
    credentials.
    """
    mocked_post.return_value = mocked_requests_post()
    mocked_getpass.return_value = "password"
    token = client.get_token()
    assert mocked_post.return_value.status_code == 200
    assert token == 'value1'


@mock.patch.object(APIClient, "post")
@mock.patch("builtins.open", new_callable=mock.mock_open, read_data="data")
def test_upload_pdf(mocked_post, _):
    """
    Tests happy path for ``otter.generate.token.APIClient.upload_pdf``.
    """
    mocked_post.return_value = mocked_requests_post()
    client.upload_pdf_submission('123', '1', 'email', 'file.txt')
    assert mocked_post.return_value.status_code == 200


@mock.patch.object(APIClient, "post")
@mock.patch("builtins.open", new_callable=mock.mock_open, read_data="data")
def test_replace_pdf(mocked_post, _):
    """
    Tests happy path for ``otter.generate.token.APIClient.replace_pdf``.
    """
    mocked_post.return_value = mocked_requests_post()
    client.replace_pdf_submission('123', '1', 'email', 'file.txt')
    assert mocked_post.return_value.status_code == 200


@mock.patch.object(APIClient, "post")
@mock.patch("builtins.open", new_callable=mock.mock_open, read_data="data")
def test_upload_programming(mocked_post, _):
    """
    Tests happy path for ``otter.generate.token.APIClient.upload_programming_submission``.
    """
    mocked_post.return_value = mocked_requests_post()
    client.upload_programming_submission('123', '1', 'email', ['python.py', 'nb.ipynb'])
    assert mocked_post.return_value.status_code == 200


# TODO: do any of these tests even check that the *correct* request is being sent out?



"""Utilities for Otter Check"""

import hashlib
import json
import os
import requests
import tempfile
import time
import wrapt

from enum import Enum
from glob import glob
from IPython import get_ipython
from IPython.display import display, Javascript
from subprocess import run, PIPE

from .logs import EventType

from ..utils import import_or_raise, NOTEBOOK_METADATA_KEY


def save_notebook(filename, timeout=10):
    """
    Force-saves a Jupyter notebook by displaying JavaScript.

    Args:
        filename (``str``): path to notebook file being saved
        timeout (``int`` or ``float``): number of seconds to wait for save before timing-out

    Returns
        ``bool``: whether the notebook was saved successfully
    """
    if get_ipython() is not None:
        orig_mod_time = os.path.getmtime(filename)
        start = time.time()
        display(Javascript("""
            if (typeof Jupyter !== 'undefined') {
                Jupyter.notebook.save_checkpoint();
            }
            else {
                document.querySelector('[data-command="docmanager:save"]').click();
            }
        """))

        while time.time() - start < timeout:
            curr_mod_time = os.path.getmtime(filename)
            if orig_mod_time < curr_mod_time and os.path.getsize(filename) > 0:
                return True

            time.sleep(0.2)

        return False

    return True


def grade_zip_file(zip_path, nb_arcname, tests_dir):
    """
    Grade a submission zip file in a separate process and return the ``GradingResults`` object.
    """
    dill = import_or_raise("dill")

    _, results_path = tempfile.mkstemp(suffix=".pkl")

    try:
        command = [
            "python3", "-m", "otter.check.validate_export",
            "--zip-path", zip_path,
            "--nb-arcname", nb_arcname,
            "--tests-dir", tests_dir,
            "--results-path", results_path,
        ]

        # run the command
        results = run(command, stdout=PIPE, stderr=PIPE)

        if results.stderr:
            raise RuntimeError(results.stderr)

        with open(results_path, "rb") as f:
            results = dill.load(f)

        return results

    finally:
        os.remove(results_path)


class IPythonInterpreter(Enum):
    """
    An enum of different types of IPython interpreters.
    """

    class Interpreter:
        """
        A class representing a flavor of IPython interpreter.

        Contains attributes for an importable name substring (used to check which interpreter is
        running) and a display name for error messages and the like.
        """

        def __init__(self, check_str, display_name):
            self.check_str = check_str
            self.display_name = display_name

        def running(self):
            """
            Determine whether this interpreter is currently running by checking the string
            representation of the return value of ``IPython.get_ipython``.

            Returns:
                ``bool``: whether this interpreter is running
            """
            return self.check_str in str(get_ipython())

    COLAB = Interpreter("google.colab", "Google Colab")
    """the Google Colab interpreter"""

    PYOLITE = Interpreter("pyolite.", "Jupyterlite")
    """the JupyterLite interpreter"""


def incompatible_with(interpreter, throw_error = True):
    """
    Create a decorator indicating that a method is incompatible with a specific interpreter.
    """
    @wrapt.decorator
    def incompatible(wrapped, self, args, kwargs):
        """
        A decorator that raises an error or performs no action (depending on ``throw_error``) if the
        wrapped function is called in an environment running on the specified interpreter.
        """
        if self._interpreter is interpreter:
            if throw_error:
                raise RuntimeError(f"This method is not compatible with {interpreter.value.display_name}")
            else:
                return
        return wrapped(*args, **kwargs)

    return incompatible


@wrapt.decorator
def grading_mode_disabled(wrapped, self, args, kwargs):
    """
    A decorator that returns without calling the wrapped function if the ``Notebook`` grading mode
    is enabled.
    """
    if type(self)._grading_mode:
        return
    return wrapped(*args, **kwargs)


# TODO: use this class to replace tuple return values in logs_event
# class LoggedEventReturnValue:
#     def __init__(self, return_value, results=None, shelve_env=None):
#         self.return_value = return_value
#         self.results = results
#         self.shelve_env = shelve_env


def logs_event(event_type):
    """
    A decorator that ensures each call is logged in the Otter log with type ``event_type``.

    Events logging a ``EventType.CHECK`` should return a 3-tuple of the question name, the
    ``TestFile`` and an environment to shelve. All other methods should just return their 
    default return value, which will be logged.
    """
    @wrapt.decorator        
    def event_logger(wrapped, self, args, kwargs):
        """
        Runs a method, catching any errors and logging the call. Returns the return value
        of the function, unless ``EventType.CHECK`` is used, in which case the return value
        is assumed to be a 3-tuple and the second value in the tuple is returned.
        """
        try:
            if event_type == EventType.CHECK:
                question, results, shelve_env = wrapped(*args, **kwargs)

            else:
                results = wrapped(*args, **kwargs)
                shelve_env = {}
                question = None

        except Exception as e:
            self._log_event(event_type, success=False, error=e)
            raise e

        else:
            self._log_event(event_type, results=results, question=question, shelve_env=shelve_env)

        return results

    return event_logger


def list_available_tests(tests_dir, nb_path):
    """
    Get a list of available questions by searching the tests directory (if present) or the notebook
    metadata.

    Args:
        tests_dir (``str``): the path to the tests directory
        nb_path (``str``): the path to the notebook

    Returns:
        ``list[str]``: the sorted list of question names
    """
    get_stem = lambda p: os.path.splitext(os.path.basename(p))[0]

    if tests_dir and os.path.isdir(tests_dir):
        tests = [get_stem(file) for file in glob(os.path.join(tests_dir, "*.py")) \
            if file != "__init__.py"]

    else:
        if nb_path is None:
            raise ValueError("Tests directory does not exist and no notebook path provided")

        with open(nb_path) as f:
            nb = json.load(f)

        tests = list(nb["metadata"][NOTEBOOK_METADATA_KEY]["tests"].keys())

    return sorted(tests)


def resolve_test_info(tests_dir, nb_path, tests_url_prefix, question):
    """
    Determine the test path and test name.

    If ``tests_url_prefix`` is specified, the test file is downloaded from the URL
    ``{tests_url_prefix}/{question}.py`` and saved to the file ``{tests_dir}/{question}.py``. If
    ``tests_dir`` does not already exist, it is created.

    Args:
        tests_dir (``str``): the path to the directory of tests
        nb_path (``str``): the path to the notebook
        tests_url_prefix (``str | None``): the prefix of a URL to the test file
        question (``str``): the question name

    Returns:
        ``tuple[str, str]``: the test path and test name
    """
    if tests_url_prefix is not None:
        if not IPythonInterpreter.PYOLITE.value.running():
            raise ValueError("Downloading test files from URLs is only supported on JupyterLite")

        pyodide = import_or_raise("pyodide")

        test_url = f"{tests_url_prefix}{'/' if not tests_url_prefix.endswith('/') else ''}{question}.py"
        text = pyodide.open_url(test_url).getvalue()

        os.makedirs(tests_dir, exist_ok=True)

        test_path = os.path.join(tests_dir, f"{question}.py")
        test_name = None

        with open(test_path, "w+") as f:
            f.write(text)

    elif tests_dir and os.path.isdir(tests_dir):
        if not os.path.isfile(os.path.join(tests_dir, question + ".py")):
            raise FileNotFoundError(f"Test {question} does not exist")

        test_path = os.path.join(tests_dir, question + ".py")
        test_name = None

    else:
        if nb_path is None:
            raise ValueError("Tests directory does not exist and no notebook path provided")

        test_path = nb_path
        test_name = question

    return test_path, test_name



"""Exception-based test files"""

import inspect
import pathlib

from dataclasses import replace
from functools import lru_cache
from textwrap import indent

from .abstract_test import TestCase, TestCaseResult, TestFile


class test_case:
    """
    A test case function decorator for Otter's exception-based test file format.

    Holds metadata for the test cases, as well as the test case function itself, and handles calling
    the function.

    Args:
        name (``str``, optional): the name of test case
        points (numeric, optional): the point value of the test case, if applicable
        hidden (``bool``, optional): whether the test case is hidden
        success_message (``str``, optional): a message to display to students if the test case passes
        failure_message (``str``, optional): a message to display to students if the test case fails

    Attributes:
        name (``str``): the name of test case
        points (numeric): the point value of the test case, if applicable
        hidden (``bool``): whether the test case is hidden
        success_message (``str``): a message to display to students if the test case passes
        failure_message (``str``): a message to display to students if the test case fails
        test_func (callable): the test case function being decorated
    """

    def __init__(self, name=None, points=None, hidden=False, success_message=None, failure_message=None):
        self.name = name
        self.points = points
        self.hidden = hidden
        self.success_message = success_message
        self.failure_message = failure_message
        self.test_func = None

    def __call__(self, test_func):
        """
        Wrap a test case function as a decorator.
        """
        self.test_func = test_func
        return self

    def to_dataclass(self):
        """
        Convert this test case to a ``TestCase`` named tuple for use in Otter's test file internals.

        Returns:
            ``otter.test_files.abstract_test.TestCase``: the test case named tuple
        """
        return TestCase(name=self.name, body=self, hidden=self.hidden, points=self.points, 
            success_message=self.success_message, failure_message=self.failure_message)

    def _get_func_params(self):
        """
        Get the list of parameters expected by the decorated test case function.

        Returns:
            ``list[str]``: the function argument names
        """
        return list(inspect.signature(self.test_func).parameters.keys())

    def call_func(self, global_environment):
        """
        Call the underlying test case function, passig in parameters from the global environment.

        If the signature of ``self.test_func`` contains a parameter called ``env``, the environment
        is passed in. For all other parameters, that value from the global environment is passed in, 
        defaulting to ``None`` if the key is not present. Thus, a function with the signature

        .. code-block:: python

            foo(env, func_a, func_b, obj)

        would get called as

        .. code-block:: python

            foo(
                env = global_environment, 
                func_a = global_environment.get("func_a"), 
                func_b = global_environment.get("func_b"),
                obj = global_environment.get("obj"),
            )

        Returns:
            ``object``: the return value of the test case function
        """
        args = self._get_func_params()
        call_kwargs = {arg: (global_environment if arg == "env" else \
                global_environment.get(arg, None)) for arg in args}
        return self.test_func(**call_kwargs)


class ExceptionTestFile(TestFile):
    """
    A single exception-based test file for Otter.

    Args:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``, optional): the point value of this test, defaults to 1
        all_or_nothing (``bool``, optional): whether the test should be graded all-or-nothing across
            cases

    Attributes:
        name (``str``): the name of test file
        path (``str``): the path to the test file
        test_cases (``list`` of ``TestCase``): a list of parsed tests to be run
        value (``int``): the point value of this test, defaults to 1
        all_or_nothing (``bool``): whether the test should be graded all-or-nothing across
            cases
        passed_all (``bool``): whether all of the test cases were passed
        test_case_results (``list`` of ``TestCaseResult``): a list of results for the test cases in
            ``test_cases``
        grade (``float``): the percentage of ``points`` earned for this test file as a decimal
        source (``str``): the test file contents
    """

    source = None

    @property
    @lru_cache(1)
    def source_lines(self):
        """
        The lines of ``self.source``
        """
        return self.source.split("\n")

    def _generate_error_message(self, excp, context=0):
        """
        Generate an error message including the line that errored from ``self.source``.

        Args:
            excp (``Exception``): the exception to generate the message for
            context (``int``, optional): a number of extra lines of context to include from above
                and below the line that caused the exception

        Returns:
            ``str``: the error message
        """
        line_idx = excp.__traceback__.tb_next.tb_next.tb_lineno - 1
        l, h = max(0, line_idx - context), min(len(self.source_lines), line_idx + context + 1)
        lines = indent("\n".join(self.source_lines[l:h]), "  ")
        err_msg = str(excp).strip("'")
        return f"Error at line {line_idx + 1} in test {self.name}:\n{lines}\n{type(excp).__name__}: {err_msg}"

    def run(self, global_environment):
        """
        Run the test cases against ``global_environment``, saving the results in 
        ``self.test_case_results``.

        Arguments:
            ``global_environment`` (``dict``): result of executing a Python notebook/script
        """
        test_case_results = []
        for tc in self.test_cases:
            test_case = tc.body
            passed, message = True, "Test case passed!"
            try:
                test_case.call_func(global_environment)
            except Exception as e:
                passed, message = False, self._generate_error_message(e)

            test_case_results.append(TestCaseResult(test_case=tc, message=message, passed=passed))

        self.test_case_results = test_case_results

    @staticmethod
    def _compile_string(s, path="<string>"):
        """
        Compile a string for execution.

        Args:
            s (``str``): the string to compile
            path (``str``, optional): the path to the test file

        Returns:
            ``code``: the compiled code of the file
        """
        return compile(s, path, "exec")

    @classmethod
    def _from_compiled_code(cls, code, path=""):
        """
        Parse a compiled exception-based test file and return an ``ExceptionTestFile``.

        Args:
            code (``code``): the compiled code of the test file
            path (``str``): the path to the test file

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        env = {}
        exec(code, env)

        if "name" not in env:
            raise ValueError(f"Test file {path} does not define 'name'")

        name = env["name"]
        points = env.get("points", None)
        test_cases = []
        for _, v in env.items():
            if isinstance(v, test_case):
                tc = v.to_dataclass()
                if tc.name is None:
                    tc = replace(tc, name=f"{name} - {len(test_cases) + 1}")
                test_cases.append(tc)

        test_cases = cls.resolve_test_file_points(points, test_cases)

        path = str(pathlib.Path(path).as_posix())
        return cls(name, path, test_cases, all_or_nothing=False)

    @classmethod
    def from_string(cls, s, path="<string>"):
        """
        Parse an exception-based test file as a string and return an ``ExceptionTestFile``.

        Args:
            s (``str``): the test file
            path (``str``, optional): the path to the test file

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        code = cls._compile_string(s, path=path)
        instc = cls._from_compiled_code(s, path=path)
        instc.source = s
        return instc

    @classmethod
    def from_file(cls, path):
        """
        Parse an exception-based test file and return an ``ExceptionTestFile``.

        Args:
            path (``str``): the path to the test file

        Returns:
            ``ExceptionTestFile``: the new ``ExceptionTestFile`` object created from the given file
        """
        with open(path) as f:
            source = f.read()
        return cls.from_string(source, path=path)



"""Tests for ``otter.assign``"""

import nbformat as nbf
import os
import pathlib
import pytest
import shutil

from glob import glob
from unittest import mock

from otter.assign import main as assign
from otter.assign.assignment import Assignment
from otter.assign.question_config import QuestionConfig
from otter.assign.tests_manager import AssignmentTestsManager, TestCase
from otter.generate.token import APIClient
from otter.utils import nullcontext

from .utils import assert_dirs_equal, dump_yaml, TestFileManager, unzip_to_temp


# prevent pytest from thinking TestCase is a testing class
TestCase.__test__ = False


FILE_MANAGER = TestFileManager("test/test-assign")


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    """
    Removes assign output
    """
    yield
    if cleanup_enabled and os.path.exists(FILE_MANAGER.get_path("output")):
        shutil.rmtree(FILE_MANAGER.get_path("output"))


@pytest.fixture(autouse=True)
def disable_pdf_generation(pdfs_enabled):
    if not pdfs_enabled:
        def create_fake_pdf(src, dest, **kwargs):
            if dest is None:
                dest = f"{pathlib.Path(src).stem}.pdf"

            open(dest, "wb+").close()
            return mock.DEFAULT

        cm1 = mock.patch("otter.assign.export_notebook", side_effect=create_fake_pdf)
        cm2 = mock.patch("otter.assign.v0.export_notebook", side_effect=create_fake_pdf)
        cm3 = mock.patch("otter.assign.knit_rmd_file", side_effect=create_fake_pdf)

    else:
        cm1, cm2, cm3 = nullcontext(), nullcontext(), nullcontext()

    with cm1, cm2, cm3:
        yield


def check_gradescope_zipfile(path, correct_dir_path):
    """
    """
    with unzip_to_temp(path) as unzipped_dir:
        assert_dirs_equal(unzipped_dir, correct_dir_path)


def assign_and_check_output(nb_path, correct_dir, assign_kwargs={}, assert_dirs_equal_kwargs={}):
    """
    """
    output_path = FILE_MANAGER.get_path("output")
    assign(nb_path, output_path, **assign_kwargs)
    assert_dirs_equal(output_path, correct_dir, **assert_dirs_equal_kwargs)


# TODO: refactor existing tests to use this
@pytest.fixture
def generate_master_notebook(cleanup_enabled):
    """
    Yields a function that generates a master notebook file with a specific assignment config and
    returns the path at which the notebook was written. This fixture also deletes the notebook file
    during cleanup.
    """
    nb_path = FILE_MANAGER.get_path("master.ipynb")

    def generate_notebook(assignment_config):
        if os.path.exists(nb_path):
            raise RuntimeError(f"{nb_path} already exists")
        nb = nbf.read(FILE_MANAGER.get_path("master-skeleton.ipynb"), as_version=nbf.NO_CONVERT)
        config_cell = nbf.v4.new_raw_cell("# ASSIGNMENT CONFIG\n" + dump_yaml(assignment_config))
        nb.cells.insert(0, config_cell)

        nbf.write(nb, nb_path)

        return nb_path

    yield generate_notebook

    if cleanup_enabled:
        os.remove(nb_path)


def test_convert_example():
    """
    Checks that otter assign filters and outputs correctly
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("example.ipynb"), 
        FILE_MANAGER.get_path("example-correct"), 
        assign_kwargs=dict(no_run_tests=True),
        assert_dirs_equal_kwargs=dict(variable_path_exts=[".zip"]),
    )

    # check gradescope zip file
    check_gradescope_zipfile(
        glob(FILE_MANAGER.get_path("output/autograder/*.zip"))[0], 
        FILE_MANAGER.get_path("example-autograder-correct"),
    )


def test_exception_example():
    """
    Checks that otter assign filters and outputs correctly
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("exception-example.ipynb"), 
        FILE_MANAGER.get_path("exception-correct"), 
        assign_kwargs=dict(no_run_tests=True),
    )


def test_otter_example():
    """
    Checks that otter assign filters and outputs correctly, as well as creates a correct .otter file
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("generate-otter.ipynb"), 
        FILE_MANAGER.get_path("otter-correct"),
        assign_kwargs=dict(v0=True),
    )


def test_pdf_example():
    """
    Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file
    along with PDFs
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("generate-pdf.ipynb"),
        FILE_MANAGER.get_path("pdf-correct"),
        assign_kwargs=dict(no_run_tests=True, v0=True),
        assert_dirs_equal_kwargs=dict(ignore_ext=[".pdf"], variable_path_exts=[".zip"]),
    )


@mock.patch.object(APIClient, "get_token")
def test_gradescope_example(mocked_client):
    """
    Checks that otter assign filters and outputs correctly, as well as creates a correct .zip file
    along with PDFs. Additionally, includes testing Gradescope integration.
    """
    # set a return value that does not match the token in the notebook, so we'll know if APIClient
    # is called
    mocked_client.return_value = 'token'

    assign_and_check_output(
        FILE_MANAGER.get_path("generate-gradescope.ipynb"),
        FILE_MANAGER.get_path("gs-correct"),
        assign_kwargs=dict(no_run_tests=True, v0=True),
        assert_dirs_equal_kwargs=dict(ignore_ext=[".pdf"], variable_path_exts=[".zip"]),
    )

    # check gradescope zip file
    check_gradescope_zipfile(
        glob(FILE_MANAGER.get_path("output/autograder/*.zip"))[0], 
        FILE_MANAGER.get_path("gs-autograder-correct"),
    )


def test_r_example():
    """
    Checks that otter assign works for R notebooks correctly
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("r-example.ipynb"),
        FILE_MANAGER.get_path("r-correct"),
        assert_dirs_equal_kwargs=dict(ignore_ext=[".pdf"], variable_path_exts=[".zip"]),
    )


def test_rmd_example():
    """
    Checks that otter assign works for Rmd files
    """
    assign_and_check_output(
        FILE_MANAGER.get_path("rmd-example.Rmd"),
        FILE_MANAGER.get_path("rmd-correct"),
        assert_dirs_equal_kwargs=dict(
            ignore_ext=[".pdf"],
            ignore_dirs=["rmd-example-sol_files"],
            variable_path_exts=[".zip"],
        ),
    )

    # check gradescope zip file
    check_gradescope_zipfile(
        glob(FILE_MANAGER.get_path("output/autograder/*.zip"))[0], 
        FILE_MANAGER.get_path("rmd-autograder-correct"),
    )


def test_point_value_rounding():
    """
    Tests that point values are rounded appropriately.
    """
    question = QuestionConfig({"name": "q1", "points": None, "manual": False})
    tests_mgr = AssignmentTestsManager(Assignment())
    for _ in range(11):
        tests_mgr._add_test_case(question, TestCase("", "", False, 4 / 11, "", ""))

    # sum(4 / 11 for _ in range(11)) evaluates to 4.000000000000001 in Python, so this will
    # check that the per-test-case point values are correctly rounded.
    points = tests_mgr.determine_question_point_value(question)
    assert points == 4


def test_jupyterlite(generate_master_notebook):
    """
    """
    master_nb_path = generate_master_notebook({
        "runs_on": "jupyterlite",
        "tests": {"url_prefix": "https://domain.tld/tests/"}
    })
    assign_and_check_output(
        master_nb_path,
        FILE_MANAGER.get_path("jupyterlite-correct")
    )



"""Tests for ``otter.Notebook``"""

import datetime as dt
import os
import pytest

from glob import glob
from textwrap import dedent
from unittest import mock

from otter import Notebook
from otter.check.notebook import _OTTER_LOG_FILENAME, _ZIP_NAME_FILENAME

from .utils import TestFileManager


FILE_MANAGER = TestFileManager("test/test-notebook")
TESTS_DIR = FILE_MANAGER.get_path("tests")
TESTS_GLOB = glob(FILE_MANAGER.get_path("tests/*.py"))


def square(x):
    return x ** 2


def negate(x):
    return not x


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):  # TODO: refactor this and similar to use delete_paths
    yield
    if cleanup_enabled and os.path.isfile(_OTTER_LOG_FILENAME):
        os.remove(_OTTER_LOG_FILENAME)


def test_check():
    """
    Checks that the cor checking behavior of ``otter.Notebook.check`` works correctly.
    """
    grader = Notebook(tests_dir=TESTS_DIR)  # TODO: move to fixture?

    def square(x):
        return x ** 2

    def negate(x):
        return not x

    global_env = {
        "square": square,
        "negate": negate,
    }

    expected_reprs = {
        "q1": 'q1 results: All test cases passed!',
        "q2": dedent("""\
            q2 results:
                q2 - 1 result:
                    Trying:
                        1 == 1
                    Expecting:
                        False
                    **********************************************************************
                    Line 2, in q2 0
                    Failed example:
                        1 == 1
                    Expected:
                        False
                    Got:
                        True

                q2 - 2 result:
                    Test case passed!
        """).strip(),
        "q3": 'q3 results: All test cases passed!',
        "q4": 'q4 results: All test cases passed!',
        "q5": 'q5 results: All test cases passed!',
    }

    for q_path in TESTS_GLOB:
        q = os.path.split(q_path)[1][:-3]
        result = grader.check(q, global_env=global_env)
        if q != "q2":
            assert result.grade == 1, "Test {} failed".format(q)
        else:
            assert result.grade == 0, "Test {} passed".format(q)

        assert repr(result) == expected_reprs[q]

        # check with no global_env
        result = grader.check(q)
        if q != "q2":
            assert result.grade == 1, "Test {} failed".format(q)
        else:
            assert result.grade == 0, "Test {} passed".format(q)


def test_to_pdf_with_nb_path():
    """
    Checks for existence of notebook PDF
    This test is the general use case WITH a specified notebook path
    """
    nb_path = "foo.ipynb"
    grader = Notebook(tests_dir=TESTS_DIR)
    with mock.patch.object(grader, "_resolve_nb_path") as mocked_resolve, \
            mock.patch("otter.check.notebook.export_notebook") as mocked_export:
        mocked_resolve.return_value = nb_path

        grader.to_pdf(filtering=False)
        mocked_export.assert_called_once_with(nb_path, filtering=False, pagebreaks=True)

        # TODO: test display_link


def test_export():
    """
    Checks export contents for existence of PDF and equality of zip
    """
    timestmap = dt.datetime(2022, 1, 3, 12, 12, 12, 1212)
    grader = Notebook(tests_dir=TESTS_DIR)

    with mock.patch.object(grader, "_resolve_nb_path") as mocked_resolve, \
            mock.patch("builtins.open", mock.mock_open(read_data="{}")) as mocked_open, \
            mock.patch("otter.check.notebook.dt") as mocked_dt, \
            mock.patch("otter.check.notebook.zipfile.ZipFile") as mocked_zf, \
            mock.patch("otter.check.notebook.export_notebook") as mocked_export, \
            open(_OTTER_LOG_FILENAME, mode="wb+"):
        mocked_resolve.return_value = "foo.ipynb"
        mocked_dt.datetime.now.return_value = timestmap

        grader.export(pdf=False)

        zip_name = f"foo_{timestmap.strftime('%Y_%m_%dT%H_%M_%S_%f')}.zip"
        mocked_zf.assert_called_once_with(zip_name, mode="w")
        mocked_zf.return_value.write.assert_any_call(mocked_resolve.return_value)
        mocked_zf.return_value.write.assert_any_call(_OTTER_LOG_FILENAME)
        mocked_export.assert_not_called()
        mocked_zf.return_value.writestr.assert_called_with(_ZIP_NAME_FILENAME, zip_name)

    # TODO: test with pdf
    # TODO: test force_save
    # TODO: test run_tests
    # TODO: test display_link


def test_colab():
    """
    Checks that the ``Notebook`` class correctly disables methods on Google Colab.
    """
    with mock.patch("otter.check.notebook.os.path.isdir") as mocked_isdir:
        mocked_isdir.return_value = False
        with pytest.raises(ValueError, match=f"Tests directory {TESTS_DIR} does not exist"):
            grader = Notebook(tests_dir=TESTS_DIR, colab=True)

        mocked_isdir.assert_called_once_with(TESTS_DIR)

        mocked_isdir.return_value = True
        grader = Notebook(tests_dir=TESTS_DIR, colab=True)

        # check for appropriate errors
        with pytest.raises(RuntimeError, match="This method is not compatible with Google Colab"):
            grader.run_plugin()

        with pytest.raises(RuntimeError, match="This method is not compatible with Google Colab"):
            grader.to_pdf()

        with pytest.raises(RuntimeError, match="This method is not compatible with Google Colab"):
            grader.add_plugin_files()

        with pytest.raises(RuntimeError, match="This method is not compatible with Google Colab"):
            grader.export()


def test_jupyterlite():
    """
    Checks that the ``Notebook`` class correctly disables methods on Google Colab.
    """
    tests_url_prefix = "https://domain.tld/"
    grader = Notebook(tests_url_prefix=tests_url_prefix, jupyterlite=True)

    # check for appropriate errors
    with mock.patch("otter.check.notebook.LogEntry") as mocked_event:
        grader._log_event()
        mocked_event.assert_not_called()

    with mock.patch("otter.check.utils.import_or_raise") as mocked_import, \
            mock.patch("otter.check.utils.os") as mocked_os, \
            mock.patch("otter.check.utils.open", mock.mock_open()) as mocked_open, \
            mock.patch("otter.check.utils.IPythonInterpreter") as mocked_interp:
        mocked_interp.PYOLITE.value.running.return_value = True
        mocked_os.path.join.return_value = FILE_MANAGER.get_path("tests/q1.py")

        grader.check("q1")

        mocked_import.assert_called_with("pyodide")
        mocked_pyodide = mocked_import.return_value
        mocked_pyodide.open_url.assert_called_with(f"{tests_url_prefix}q1.py")
        mocked_os.makedirs.assert_called_with("./tests", exist_ok=True)
        mocked_open.assert_called_with(mocked_os.path.join.return_value, "w+")


def test_grading_mode():
    """
    Check that a call to a grading-mode-disabled method is not executed.
    """
    with mock.patch.object(Notebook, "_resolve_nb_path") as mocked_resolve, \
            Notebook.grading_mode(tests_dir="foo"):
        grader = Notebook(tests_dir=TESTS_DIR)
        grader.export()

        # TODO: find a better way of doing this than accessing a private field
        assert grader._path == "foo"

        # if export is called, this method would be called first
        mocked_resolve.assert_not_called()


# TODO: tests for force_save on export and to_pdf
# TODO: test _resolve_nb_path
# TODO: tests for event logging and other things in otter.check.utils



"""Tests for ``otter.export``"""

# NOTES:
# - tests do not check for PDF equality

import filecmp
import nbconvert
import nbformat
import os
import pytest

from glob import glob
from unittest import mock

from otter.export import main as export
from otter.export.exporters.base_exporter import BaseExporter
from otter.utils import nullcontext

from .utils import TestFileManager


FILE_MANAGER = TestFileManager("test/test-export")


@pytest.fixture(autouse=True)
def cleanup_output(cleanup_enabled):
    """
    Removes export output
    """
    yield
    if cleanup_enabled:
        for file in glob(FILE_MANAGER.get_path("*.pdf")) + glob(FILE_MANAGER.get_path("*.tex")) + \
                [FILE_MANAGER.get_path("output.ipynb")]:
            if os.path.exists(file):
                os.remove(file)


@pytest.fixture(autouse=True)
def disable_pdf_generation(pdfs_enabled):
    if not pdfs_enabled:
        def create_fake_pdf(exporter, nb):
            contents = "pdf contents"
            if isinstance(exporter, nbconvert.PDFExporter):
                contents = contents.encode("utf-8")
            return contents, {}

        cm = mock.patch("otter.export.exporters.via_html.nbconvert.export", side_effect=create_fake_pdf)
        cm2 = mock.patch("otter.export.exporters.via_latex.nbconvert.export", side_effect=create_fake_pdf)

    else:
        cm, cm2 = nullcontext(), nullcontext()

    with cm, cm2:
        yield


def run_export(notebook_path, **kwargs):
    export(notebook_path, **kwargs)


def test_success_HTML():
    """
    Tests a successful export with filtering and no pagebreaks
    """
    test_file = "successful-html-test"
    run_export(
        FILE_MANAGER.get_path(f"{test_file}.ipynb"), filtering=True, save=True, exporter="latex")

    # check existence of pdf and tex
    FILE_MANAGER.assert_path_exists(FILE_MANAGER.get_path(f"{test_file}.pdf"), dir_okay=False)
    FILE_MANAGER.assert_path_exists(FILE_MANAGER.get_path(f"{test_file}.tex"), dir_okay=False)


def test_success_pagebreak():
    """
    Tests a successful filter with pagebreaks
    """
    test_file = "success-pagebreak-test"
    run_export(
        FILE_MANAGER.get_path(f"{test_file}.ipynb"), 
        filtering=True, 
        exporter="latex", 
        pagebreaks=True, 
        save=True,
    )

    # check existence of pdf and tex
    FILE_MANAGER.assert_path_exists(FILE_MANAGER.get_path(f"{test_file}.pdf"), dir_okay=False)
    FILE_MANAGER.assert_path_exists(FILE_MANAGER.get_path(f"{test_file}.tex"), dir_okay=False)


def test_no_close():
    """
    Tests a filtered export without a closing comment
    """
    test_file = "no-close-tag-test"
    run_export(
        FILE_MANAGER.get_path(f"{test_file}.ipynb"), 
        filtering=True, 
        exporter="latex", 
        pagebreaks=True, 
        save=True,
    )

    # check existence of pdf and tex
    FILE_MANAGER.assert_path_exists(FILE_MANAGER.get_path(f"{test_file}.pdf"), dir_okay=False)
    FILE_MANAGER.assert_path_exists(FILE_MANAGER.get_path(f"{test_file}.tex"), dir_okay=False)


def test_load_notebook():
    """
    Tests a successful load_notebook
    """
    test_file = "successful-html-test"
    node = BaseExporter.load_notebook(FILE_MANAGER.get_path(f"{test_file}.ipynb"), filtering=True)

    nbformat.write(node, FILE_MANAGER.get_path("output.ipynb"))

    # check file contents
    assert filecmp.cmp(FILE_MANAGER.get_path("output.ipynb"), FILE_MANAGER.get_path(f"correct/{test_file}.ipynb"))



"""Assignment creation tool for Otter-Grader"""

import json
import os
import pathlib
import warnings

from .assignment import Assignment
from .output import write_output_directories
from .utils import run_tests, write_otter_config_file, run_generate_autograder

from ..export import export_notebook, WkhtmltopdfNotFoundError
from ..plugins import PluginCollection
from ..utils import chdir, get_relpath, knit_rmd_file, loggers


LOGGER = loggers.get_logger(__name__)


def main(master, result, *, no_pdfs=False, no_run_tests=False, username=None, password=None, 
         debug=False, v0=False):
    """
    Runs Otter Assign on a master notebook.

    Args:
        master (``str``): path to master notebook
        result (``str``): path to result directory
        no_pdfs (``bool``): whether to ignore any configurations indicating PDF generation for this run
        no_run_tests (``bool``): prevents Otter tests from being automatically run on the solutions 
            notebook
        username (``str``): a username for Gradescope for generating a token
        password (``str``): a password for Gradescope for generating a token
        debug (``bool``): whether to run in debug mode (without ignoring errors during testing)
        v0 (``bool``): whether to use Otter Assign Format v0 instead of v1
    """
    if v0:
        warnings.warn(
            "The Otter Assign v0 format is now deprecated and will be removed in Otter v5.",
            FutureWarning)

        from .v0 import main as v0_main
        return v0_main(master, result, no_pdfs=no_pdfs, no_run_tests=no_run_tests, username=username, 
            password=password, debug=debug)

    LOGGER.debug(f"User-specified master path: {master}")
    LOGGER.debug(f"User-specified result path: {result}")
    master, result = pathlib.Path(os.path.abspath(master)), pathlib.Path(os.path.abspath(result))

    assignment = Assignment()

    result = get_relpath(master.parent, result)

    assignment.master, assignment.result = master, result
    LOGGER.debug(f"Normalized master path: {master}")
    LOGGER.debug(f"Normalized result path: {result}")

    with chdir(master.parent):
        LOGGER.info("Generating views")
        write_output_directories(assignment)

        # update seed variables
        if assignment.seed.variable:
            LOGGER.debug("Processing seed dict")
            if assignment.generate_enabled:
                LOGGER.debug("Otter Generate configuration found while processing seed dict")
                assignment.generate.seed = assignment.seed.autograder_value
                assignment.generate.seed_variable = assignment.seed.variable
                LOGGER.debug("Added seed information to assignment.generate")

        # check that we have a seed if needed
        if assignment.seed_required:
            LOGGER.debug("Assignment seed is required")
            if assignment.generate_enabled and \
                    not isinstance(assignment.generate.seed, int):
                raise RuntimeError("Seeding cell found but no or invalid seed provided")

        plugins, pc = assignment.plugins, None
        if plugins:
            LOGGER.debug("Processing plugins")
            pc = PluginCollection(plugins, "", {})
            pc.run("during_assign", assignment)
            if assignment.generate_enabled:
                LOGGER.debug("Adding plugin configurations to Otter Generate configuration")
                assignment.generate.plugins.extend(plugins)

        # generate Gradescope autograder zipfile
        if assignment.generate_enabled:
            LOGGER.info("Generating autograder zipfile")
            run_generate_autograder(assignment, username, password, plugin_collection=pc)

        # generate PDF of solutions
        if assignment.solutions_pdf and not no_pdfs:
            LOGGER.info("Generating solutions PDF")
            filtering = assignment.solutions_pdf == 'filtered'

            src = os.path.abspath(str(assignment.get_ag_path(master.name)))
            dst = os.path.abspath(str(assignment.get_ag_path(master.stem + '-sol.pdf')))

            if not assignment.is_rmd:
                LOGGER.debug(f"Exporting {src} as notebook to {dst}")
                try:
                    LOGGER.debug("Attempting PDF via HTML export")
                    export_notebook(
                        src,
                        dest=dst,
                        filtering=filtering,
                        pagebreaks=filtering,
                        exporter_type="html",
                    )
                    LOGGER.debug("PDF via HTML export successful")

                except WkhtmltopdfNotFoundError:
                    LOGGER.debug("PDF via HTML export failed; attempting PDF via LaTeX export")
                    export_notebook(
                        src,
                        dest=dst,
                        filtering=filtering,
                        pagebreaks=filtering,
                    )
                    LOGGER.debug("PDF via LaTeX export successful")

            else:
                LOGGER.debug(f"Knitting {src} to {dst}")
                if filtering:
                    raise ValueError("Filtering is not supported with RMarkdown assignments")

                knit_rmd_file(src, dst)

        # generate a tempalte PDF for Gradescope
        if assignment.template_pdf and not no_pdfs:
            LOGGER.info("Generating template PDF")

            src = os.path.abspath(str(assignment.get_ag_path( master.name)))
            dst = os.path.abspath(str(assignment.get_ag_path(master.stem + '-template.pdf')))

            if not assignment.is_rmd:
                LOGGER.debug("Attempting PDF via LaTeX export")
                export_notebook(
                    src,
                    dest=dst, 
                    filtering=True, 
                    pagebreaks=True, 
                    exporter_type="latex",
                )
                LOGGER.debug("PDF via LaTeX export successful")

            else:
                raise ValueError(f"Filtering is not supported with RMarkdown assignments; use " + \
                    "solutions_pdf to generate a Gradescope template instead.")

        # generate the .otter file if needed
        if not assignment.is_rmd and assignment.save_environment:
            LOGGER.debug("Processing environment serialization configuration")
            if assignment.is_r:
                warnings.warn(
                    "Otter Service and serialized environments are unsupported with R, "
                    "configurations ignored")
            else:
                write_otter_config_file(assignment)

        # run tests on autograder notebook
        if assignment.run_tests and not no_run_tests and assignment.is_python:
            LOGGER.info("Running tests against the solutions notebook")

            seed = assignment.generate.seed if assignment.generate_enabled else None
            LOGGER.debug(f"Resolved seed for running tests: {seed}")

            if assignment.generate_enabled:
                LOGGER.debug("Retrieving updated plugins from otter_config.json for running tests")
                test_pc = PluginCollection(
                    assignment.generate.plugins, assignment.ag_notebook_path, {})

            else:
                LOGGER.debug("Using pre-configured plugins for running tests")
                test_pc = pc

            run_tests(
                assignment.get_ag_path(master.name),
                debug=debug,
                seed=seed,
                plugin_collection=test_pc,
            )

            LOGGER.info("All autograder tests passed.")



"""Cell generators for R notebooks"""

import nbformat

from otter.assign.feature_toggle import FeatureToggle

from ..cell_factory import CellFactory
from ..utils import lock


class RCellFactory(CellFactory):
    """
    A cell factory for R assignments.
    """

    def create_init_cells(self):
        return []

    def create_check_cells(self, question):
        cell = nbformat.v4.new_code_cell()
        cell.source = ['. = ottr::check("tests/{}.R")'.format(question.name)]
        lock(cell)
        return [cell]

    def create_check_all_cells(self):
        return []

    def create_export_cells(self):
        if not self.assignment.export_cell:
            return []

        instructions = nbformat.v4.new_markdown_cell()
        instructions.source = "## Submission\n\nMake sure you have run all cells in your " \
            "notebook in order before running the cell below, so that all images/graphs appear " \
            "in the output. The cell below will generate a zip file for you to submit. **Please " \
            "save before exporting!**"

        if self.assignment.export_cell.instructions:
            instructions.source += '\n\n' + self.assignment.export_cell.instructions

        export = nbformat.v4.new_code_cell()
        source_lines = ["# Save your notebook first, then run this cell to export your submission."]
        source_lines.append(f'ottr::export("{self.assignment.notebook_basename}")')
        export.source = "\n".join(source_lines)

        lock(instructions)
        lock(export)

        cells = [instructions, export]
        if self.check_feature_toggle(FeatureToggle.EMPTY_MD_BOUNDARY_CELLS):
            cells.append(nbformat.v4.new_markdown_cell(" "))  # add buffer cell

        return cells
